"""MCD (Mel Cepstral Distortion) module.

Full-reference metric for evaluating TTS and voice conversion quality.
Computes the Euclidean distance between MFCC vectors of reference and
synthesised speech.

Score range: 0.0+ dB (lower = better).
  <4 dB   excellent (near-natural speech)
  4-6 dB  good
  6-8 dB  acceptable
  >8 dB   poor

Standard metric in TTS research since Kubichek (1993).

References:
    - Kubichek (1993), "Mel-Cepstral Distance Measure for Objective
      Speech Quality Assessment"
"""

import logging
from pathlib import Path
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AudioMCDModule(PipelineModule):
    name = "audio_mcd"
    description = "Mel Cepstral Distortion for TTS/VC quality (full-reference)"
    default_config = {
        "target_sr": 16000,
        "n_mfcc": 13,
        "warning_threshold": 8.0,  # dB
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.target_sr = self.config.get("target_sr", 16000)
        self.n_mfcc = self.config.get("n_mfcc", 13)
        self.warning_threshold = self.config.get("warning_threshold", 8.0)
        self._ml_available = False

    def setup(self) -> None:
        try:
            import librosa  # noqa: F401

            self._ml_available = True
            logger.info("MCD module initialised (librosa)")
        except ImportError:
            logger.warning("librosa not installed. Install with: pip install librosa")

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample
        reference = Path(reference) if not isinstance(reference, Path) else reference
        if not reference.exists():
            return sample

        try:
            ref_mfcc = self._extract_mfcc(reference)
            syn_mfcc = self._extract_mfcc(sample.path)

            if ref_mfcc is None or syn_mfcc is None:
                return sample

            # Align to shorter
            min_frames = min(ref_mfcc.shape[1], syn_mfcc.shape[1])
            if min_frames < 10:
                return sample

            ref_mfcc = ref_mfcc[:, :min_frames]
            syn_mfcc = syn_mfcc[:, :min_frames]

            # MCD: (10 * sqrt(2) / ln(10)) * mean(||ref - syn||_2)
            # Exclude c0 (energy), use c1..c_n_mfcc
            diff = ref_mfcc[1:, :] - syn_mfcc[1:, :]
            frame_dist = np.sqrt(np.sum(diff ** 2, axis=0))
            mcd = (10.0 * np.sqrt(2.0) / np.log(10.0)) * np.mean(frame_dist)

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.mcd_score = round(float(mcd), 3)

            if mcd > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"High Mel Cepstral Distortion: {mcd:.2f} dB",
                        details={"mcd": float(mcd)},
                    )
                )

        except Exception as e:
            logger.warning(f"MCD failed for {sample.path}: {e}")

        return sample

    def _extract_mfcc(self, path: Path) -> Optional[np.ndarray]:
        try:
            import librosa

            y, sr = librosa.load(str(path), sr=self.target_sr, mono=True)
            mfcc = librosa.feature.mfcc(y=y, sr=sr, n_mfcc=self.n_mfcc + 1)
            return mfcc
        except Exception as e:
            logger.debug(f"MFCC extraction failed: {e}")
            return None



import copy
import logging
import os
import os.path as osp
from os.path import join

import torch
import deepspeed
from torch.utils.data import ConcatDataset, DataLoader

from utils.optimizer import create_optimizer
from utils.scheduler import create_scheduler

logger = logging.getLogger(__name__)


def get_media_types(datasources):
    """get the media types for for all the dataloaders.

    Args:
        datasources (List): List of dataloaders or datasets.

    Returns: List. The media_types.

    """
    if isinstance(datasources[0], DataLoader):
        datasets = [dataloader.dataset for dataloader in datasources]
    else:
        datasets = datasources
    media_types = [
        dataset.datasets[0].media_type
        if isinstance(dataset, ConcatDataset)
        else dataset.media_type
        for dataset in datasets
    ]

    return media_types


def setup_model(
    config, model_cls, pretrain=False, find_unused_parameters=False, num_steps_per_epoch=-1,
):
    logger.info("Creating model")
    config = copy.deepcopy(config)

    model = model_cls(config=config, is_pretrain=pretrain)

    model = model.to(torch.device(config.device))
    if config.use_half_precision:
        if config.get('bf16', True):
            logger.info("Change to bfloat16 for model")
            model = model.to(torch.bfloat16)
        else:
            logger.info("Change to float16 for model")
            model = model.half()
    tokenizer = model.tokenizer
    model_without_ddp = model

    if hasattr(config, "deepspeed") and config.deepspeed.enable:
        optimizer_params = create_optimizer(config.optimizer, model, return_group=True)
        scheduler = None
        scaler = None
    else:
        if config.distributed:
            model = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[config.gpu],
                find_unused_parameters=find_unused_parameters,  # `False` for image-only task
            )

        optimizer = create_optimizer(config.optimizer, model)
        scheduler = create_scheduler(config.scheduler, optimizer)
        scaler = torch.cuda.amp.GradScaler(enabled=config.use_half_precision) # This is never used actually if we fixed bf16

    start_epoch = 0
    global_step = 0

    # auto resume the latest checkpoint
    if config.get("auto_resume", False):
        logger.info("Auto resuming")
        model_latest = join(config.output_dir, "ckpt_latest.pth")
        model_best = join(config.output_dir, "ckpt_best.pth")

        large_step_num = -1
        large_num = -1
        for p in os.listdir(config.output_dir):
            if 'ckpt_iter' in p:
                num = p.split('_iter')[1].split('.')[0]
                if str.isnumeric(num):
                    if int(num) > large_step_num:
                        large_step_num = int(num)
            elif 'ckpt_' in p:
                num = p.split('_')[1].split('.')[0]
                if str.isnumeric(num):
                    if int(num) > large_num:
                        large_num = int(num)
        if large_step_num != -1:
            logger.info(f"Load the latest step: {large_step_num}")
            model_latest = join(config.output_dir, f"ckpt_iter{large_step_num:02d}.pth")
        if large_num != -1 and (large_num + 1) * num_steps_per_epoch > large_step_num:
            logger.info(f"Load the latest epoch: {large_num}")
            model_latest = join(config.output_dir, f"ckpt_{large_num:02d}.pth")

        if hasattr(config, "deepspeed") and config.deepspeed.enable:
            if osp.isdir(model_latest):
                config.pretrained_path = model_latest
                config.resume = True
            elif osp.isdir(model_best):
                config.pretrained_path = model_best
                config.resume = True
            else:
                logger.info(f"Not found checkpoint in {config.output_dir}")
        else:
            if osp.isfile(model_latest):
                config.pretrained_path = model_latest
                config.resume = True
            elif osp.isfile(model_best):
                config.pretrained_path = model_best
                config.resume = True
            else:
                logger.info(f"Not found checkpoint in {config.output_dir}")

    # load pretrained model
    if hasattr(config, "deepspeed") and config.deepspeed.enable:
        logger.info('Use deepspeed to initialize model!!!')
        model = model_without_ddp
        model, optimizer, _, _ = deepspeed.initialize(
            args=config, model=model, model_parameters=optimizer_params, dist_init_required=not config.distributed,
            lr_scheduler=lambda opt: create_scheduler(config.scheduler, opt)
        )
        if osp.isdir(config.pretrained_path):
            logger.info(f"Load pretrained model from {config.pretrained_path}")
            output_dir, tag = os.path.split(config.pretrained_path)
            if config.resume:
                _, client_state = model.load_checkpoint(output_dir, tag=tag, load_module_strict=False)
                global_step = model.global_steps
                assert num_steps_per_epoch > 0, "Please provide num_steps_per_epoch"
                start_epoch = global_step // num_steps_per_epoch
            else:
                _, client_state = model.load_checkpoint(
                    output_dir, tag=tag, load_module_strict=False, 
                    load_optimizer_states=False, load_lr_scheduler_states=False,
                    load_module_only=True
                )
    else:
        if osp.isfile(config.pretrained_path):
            checkpoint = torch.load(config.pretrained_path, map_location="cpu")
            logger.info(f"Load pretrained model from {config.pretrained_path}")
            if 'model' in checkpoint.keys():
                state_dict = checkpoint["model"]
            elif 'module' in checkpoint.keys():
                state_dict = checkpoint["module"]
            else:
                state_dict = checkpoint
            # resume optimizer
            if config.resume:
                optimizer.load_state_dict(checkpoint["optimizer"])
                scheduler.load_state_dict(checkpoint["scheduler"])
                scaler.load_state_dict(checkpoint["scaler"])
                start_epoch = checkpoint["epoch"] + 1
                global_step = checkpoint["global_step"]

            msg = model_without_ddp.load_state_dict(state_dict, strict=False)
            logger.info(msg)
            logger.info(f"Loaded checkpoint from {config.pretrained_path}")
        else:
            logger.warning("No pretrained checkpoint provided, training from scratch")
    
    logger.info(f"Cuda memory after create model: {torch.cuda.memory_allocated() // 1024**2}M, Max mem: {torch.cuda.max_memory_allocated() // 1024**2}M")

    return (
        model,
        model_without_ddp,
        optimizer,
        scheduler,
        scaler,
        tokenizer,
        start_epoch,
        global_step,
    )



"""T2V-CompBench module — CVPR 2025.

Seven compositional sub-metrics for text-to-video generation evaluation:
  1. Attribute binding — does generated object have the described attribute?
  2. Object relationship — are described inter-object relations present?
  3. Action binding — does the subject perform the described action?
  4. Spatial relationship — correct spatial arrangement?
  5. Generative numeracy — correct count of objects?
  6. Scene composition — overall scene matches caption?
  7. Overall — weighted mean of sub-metrics.

Backend tiers:
  1. **YOLO+Depth+CLIP** — YOLO-World detection + Depth Anything V2 + CLIP verification
  2. **CLIP-only** — CLIP text-image matching (no detection, skip spatial/numeracy)
  3. **Heuristic** — Text parsing + basic spatial analysis

Video-only (requires caption).
"""

import logging
import re
from typing import Dict, List, Optional, Tuple

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

# Number words for numeracy parsing
WORD_TO_NUM = {
    "zero": 0, "no": 0, "one": 1, "a": 1, "an": 1, "single": 1,
    "two": 2, "couple": 2, "pair": 2,
    "three": 3, "four": 4, "five": 5, "six": 6,
    "seven": 7, "eight": 8, "nine": 9, "ten": 10,
    "several": -1, "many": -1, "multiple": -1, "few": -1,
}

# Spatial prepositions
SPATIAL_PREPS = [
    "above", "below", "left of", "right of", "behind", "in front of",
    "on top of", "under", "next to", "beside", "between", "inside",
    "outside", "near", "far from",
]

# Relationship words
RELATION_WORDS = [
    "holding", "carrying", "wearing", "riding", "sitting on",
    "standing on", "touching", "hugging", "pushing", "pulling",
    "chasing", "following", "leading", "watching",
]

# Action verbs (gerund form)
ACTION_VERBS = [
    "running", "walking", "jumping", "dancing", "swimming", "flying",
    "eating", "drinking", "sleeping", "talking", "laughing", "crying",
    "playing", "fighting", "climbing", "falling", "spinning", "turning",
    "waving", "pointing", "throwing", "catching", "kicking",
]


class T2VCompBenchModule(PipelineModule):
    name = "t2v_compbench"
    description = "T2V-CompBench compositional metrics (YOLO+Depth+CLIP / CLIP / heuristic)"
    default_config = {
        "subsample": 8,
        "enable_attribute": True,
        "enable_object_rel": True,
        "enable_action": True,
        "enable_spatial": True,
        "enable_numeracy": True,
        "enable_scene": True,
        "weights": [1, 1, 1, 1, 1, 1],
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._backend = "heuristic"
        self._yolo = None
        self._depth = None
        self._clip_model = None
        self._clip_processor = None
        self._device = "cpu"

    def setup(self) -> None:
        # Tier 1: YOLO-World + Depth Anything + CLIP
        try:
            import torch
            self._device = "cuda" if torch.cuda.is_available() else "cpu"
        except ImportError:
            self._device = "cpu"

        yolo_ok = self._try_load_yolo()
        clip_ok = self._try_load_clip()
        depth_ok = self._try_load_depth()

        if yolo_ok and clip_ok:
            self._backend = "yolo_depth" if depth_ok else "yolo_clip"
            logger.info("T2VCompBench using YOLO+CLIP%s backend", "+Depth" if depth_ok else "")
        elif clip_ok:
            self._backend = "clip"
            logger.info("T2VCompBench using CLIP-only backend")
        else:
            self._backend = "heuristic"
            logger.info("T2VCompBench using heuristic backend")

    def _try_load_yolo(self) -> bool:
        try:
            from ultralytics import YOLO
            self._yolo = YOLO("yolov8s-world.pt")
            return True
        except Exception as e:
            logger.info("YOLO-World unavailable: %s", e)
            return False

    def _try_load_clip(self) -> bool:
        try:
            from transformers import CLIPModel, CLIPProcessor
            models_dir = self.config.get("models_dir", "models")
            self._clip_model = CLIPModel.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir=models_dir,
            ).to(self._device)
            self._clip_model.eval()
            self._clip_processor = CLIPProcessor.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir=models_dir,
            )
            return True
        except Exception as e:
            logger.info("CLIP unavailable: %s", e)
            return False

    def _try_load_depth(self) -> bool:
        try:
            from transformers import pipeline
            self._depth = pipeline(
                "depth-estimation",
                model="depth-anything/Depth-Anything-V2-Small-hf",
                device=0 if self._device == "cuda" else -1,
            )
            return True
        except Exception as e:
            logger.info("Depth Anything unavailable: %s", e)
            return False

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample
        if not sample.caption or not sample.caption.text:
            return sample

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        try:
            caption = sample.caption.text
            frames = self._extract_frames(sample)
            if len(frames) < 2:
                return sample

            scores = {}

            if self._backend == "yolo_depth":
                scores = self._compute_yolo_depth(frames, caption)
            elif self._backend == "clip":
                scores = self._compute_clip_only(frames, caption)
            else:
                scores = self._compute_heuristic(frames, caption)

            # Write scores to quality metrics
            qm = sample.quality_metrics
            if self.config.get("enable_attribute", True) and "attribute" in scores:
                qm.compbench_attribute = scores["attribute"]
            if self.config.get("enable_object_rel", True) and "object_rel" in scores:
                qm.compbench_object_rel = scores["object_rel"]
            if self.config.get("enable_action", True) and "action" in scores:
                qm.compbench_action = scores["action"]
            if self.config.get("enable_spatial", True) and "spatial" in scores:
                qm.compbench_spatial = scores["spatial"]
            if self.config.get("enable_numeracy", True) and "numeracy" in scores:
                qm.compbench_numeracy = scores["numeracy"]
            if self.config.get("enable_scene", True) and "scene" in scores:
                qm.compbench_scene = scores["scene"]

            # Overall = weighted mean
            weights = self.config.get("weights", [1, 1, 1, 1, 1, 1])
            sub_keys = ["attribute", "object_rel", "action", "spatial", "numeracy", "scene"]
            valid_scores = []
            valid_weights = []
            for i, key in enumerate(sub_keys):
                if key in scores and scores[key] is not None:
                    valid_scores.append(scores[key])
                    valid_weights.append(weights[i] if i < len(weights) else 1)

            if valid_scores:
                w_sum = sum(valid_weights)
                overall = sum(s * w for s, w in zip(valid_scores, valid_weights)) / (w_sum + 1e-8)
                qm.compbench_overall = float(np.clip(overall, 0.0, 1.0))

        except Exception as e:
            logger.warning("T2VCompBench processing failed: %s", e)

        return sample

    # ------------------------------------------------------------------ #
    # Frame extraction                                                     #
    # ------------------------------------------------------------------ #

    def _extract_frames(self, sample: Sample) -> list:
        num_frames = self.config.get("subsample", 8)
        cap = cv2.VideoCapture(str(sample.path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        if total < 2:
            cap.release()
            return []
        indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]
        frames = []
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                frames.append(frame)
        cap.release()
        return frames

    # ------------------------------------------------------------------ #
    # Text parsing utilities                                               #
    # ------------------------------------------------------------------ #

    @staticmethod
    def _parse_attributes(caption: str) -> List[Tuple[str, str]]:
        """Parse 'adjective noun' pairs from caption."""
        # Simple pattern: adj + noun
        pattern = re.compile(
            r'\b(red|blue|green|yellow|black|white|large|small|big|tiny|tall|short|'
            r'old|young|new|bright|dark|round|square|long|thin|thick|heavy|light|'
            r'beautiful|ugly|colorful|shiny|smooth|rough|soft|hard|fast|slow|'
            r'wooden|metal|glass|golden|silver|furry|striped|spotted)\s+'
            r'(\w+)\b', re.IGNORECASE
        )
        return pattern.findall(caption)

    @staticmethod
    def _parse_spatial(caption: str) -> List[Tuple[str, str, str]]:
        """Parse 'X spatial_prep Y' patterns."""
        results = []
        caption_lower = caption.lower()
        for prep in SPATIAL_PREPS:
            pattern = re.compile(r'(\w+)\s+' + re.escape(prep) + r'\s+(?:the\s+)?(\w+)', re.IGNORECASE)
            for m in pattern.finditer(caption_lower):
                results.append((m.group(1), prep, m.group(2)))
        return results

    @staticmethod
    def _parse_count(caption: str) -> List[Tuple[int, str]]:
        """Parse 'number noun' patterns."""
        results = []
        number_words = set(WORD_TO_NUM.keys())
        # Match: number word/digit + optional 1-2 adjectives + noun
        pattern = re.compile(
            r'\b(' + '|'.join(re.escape(w) for w in WORD_TO_NUM) + r'|\d+)\s+(\w+(?:\s+\w+)?)\b',
            re.IGNORECASE
        )
        for m in pattern.finditer(caption):
            word = m.group(1).lower()
            rest = m.group(2).strip()
            # Take the last word as noun (skip adjective words and other number words)
            parts = rest.split()
            noun = parts[-1]
            # Skip if noun is itself a number word
            if noun.lower() in number_words:
                continue
            if word in WORD_TO_NUM:
                num = WORD_TO_NUM[word]
            else:
                try:
                    num = int(word)
                except ValueError:
                    continue
            if num >= 0:  # Skip vague counts
                results.append((num, noun))
        return results

    @staticmethod
    def _parse_actions(caption: str) -> List[Tuple[str, str]]:
        """Parse 'noun verb-ing' patterns."""
        results = []
        caption_lower = caption.lower()
        for verb in ACTION_VERBS:
            pattern = re.compile(r'(\w+)\s+' + re.escape(verb), re.IGNORECASE)
            for m in pattern.finditer(caption_lower):
                results.append((m.group(1), verb))
        return results

    @staticmethod
    def _parse_relations(caption: str) -> List[Tuple[str, str, str]]:
        """Parse 'X relation Y' patterns."""
        results = []
        caption_lower = caption.lower()
        for rel in RELATION_WORDS:
            pattern = re.compile(r'(\w+)\s+' + re.escape(rel) + r'\s+(?:a\s+|the\s+)?(\w+)', re.IGNORECASE)
            for m in pattern.finditer(caption_lower):
                results.append((m.group(1), rel, m.group(2)))
        return results

    # ------------------------------------------------------------------ #
    # CLIP similarity helper                                               #
    # ------------------------------------------------------------------ #

    def _clip_sim(self, image: np.ndarray, text: str) -> float:
        """Compute CLIP cosine similarity between image and text."""
        import torch
        from PIL import Image

        rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        pil_img = Image.fromarray(rgb)

        inputs = self._clip_processor(
            text=[text], images=pil_img, return_tensors="pt", padding=True
        ).to(self._device)

        with torch.no_grad():
            outputs = self._clip_model(**inputs)
            logits = outputs.logits_per_image
            return float(logits[0, 0].item() / 100.0)

    def _clip_sim_batch_text(self, image: np.ndarray, texts: List[str]) -> List[float]:
        """CLIP similarity of one image against multiple texts."""
        import torch
        from PIL import Image

        rgb = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        pil_img = Image.fromarray(rgb)

        inputs = self._clip_processor(
            text=texts, images=pil_img, return_tensors="pt", padding=True
        ).to(self._device)

        with torch.no_grad():
            outputs = self._clip_model(**inputs)
            sims = outputs.logits_per_image.softmax(dim=-1)[0]
            return [float(s) for s in sims]

    # ------------------------------------------------------------------ #
    # Tier 1: YOLO + Depth + CLIP                                          #
    # ------------------------------------------------------------------ #

    def _detect_objects(self, frame: np.ndarray, classes: List[str]) -> list:
        """Run YOLO-World detection for specific classes."""
        self._yolo.set_classes(classes)
        results = self._yolo(frame, verbose=False)
        detections = []
        if results and len(results) > 0:
            r = results[0]
            for box, cls_id, conf in zip(r.boxes.xyxy.cpu().numpy(),
                                          r.boxes.cls.cpu().numpy().astype(int),
                                          r.boxes.conf.cpu().numpy()):
                detections.append({
                    "class": classes[cls_id] if cls_id < len(classes) else "unknown",
                    "box": box.tolist(),
                    "conf": float(conf),
                })
        return detections

    def _compute_yolo_depth(self, frames: list, caption: str) -> Dict[str, float]:
        scores = {}

        # Attribute binding
        attrs = self._parse_attributes(caption)
        if attrs:
            attr_scores = []
            for adj, noun in attrs:
                for frame in frames[::max(1, len(frames) // 4)]:
                    dets = self._detect_objects(frame, [noun])
                    if dets:
                        # Crop detected region and verify attribute via CLIP
                        det = max(dets, key=lambda d: d["conf"])
                        x1, y1, x2, y2 = [int(c) for c in det["box"]]
                        crop = frame[max(0, y1):y2, max(0, x1):x2]
                        if crop.size > 0:
                            sim = self._clip_sim(crop, f"a {adj} {noun}")
                            attr_scores.append(sim)
            scores["attribute"] = float(np.mean(attr_scores)) if attr_scores else 0.5

        # Object relationship
        rels = self._parse_relations(caption)
        if rels:
            rel_scores = []
            for subj, rel, obj in rels:
                for frame in frames[::max(1, len(frames) // 4)]:
                    dets_subj = self._detect_objects(frame, [subj])
                    dets_obj = self._detect_objects(frame, [obj])
                    if dets_subj and dets_obj:
                        # Both objects detected — verify relation via CLIP
                        sim = self._clip_sim(frame, f"{subj} {rel} {obj}")
                        rel_scores.append(sim)
                    elif dets_subj or dets_obj:
                        rel_scores.append(0.3)
                    else:
                        rel_scores.append(0.0)
            scores["object_rel"] = float(np.mean(rel_scores)) if rel_scores else 0.5

        # Action binding
        actions = self._parse_actions(caption)
        if actions:
            action_scores = []
            for noun, verb in actions:
                frame_scores = []
                for frame in frames[::max(1, len(frames) // 3)]:
                    dets = self._detect_objects(frame, [noun])
                    if dets:
                        det = max(dets, key=lambda d: d["conf"])
                        x1, y1, x2, y2 = [int(c) for c in det["box"]]
                        crop = frame[max(0, y1):y2, max(0, x1):x2]
                        if crop.size > 0:
                            sim = self._clip_sim(crop, f"a {noun} {verb}")
                            frame_scores.append(sim)
                if frame_scores:
                    action_scores.append(float(np.mean(frame_scores)))
            scores["action"] = float(np.mean(action_scores)) if action_scores else 0.5

        # Spatial relationship
        spatials = self._parse_spatial(caption)
        if spatials:
            spatial_scores = []
            for subj, prep, obj in spatials:
                for frame in frames[::max(1, len(frames) // 4)]:
                    dets_subj = self._detect_objects(frame, [subj])
                    dets_obj = self._detect_objects(frame, [obj])
                    if dets_subj and dets_obj:
                        s_box = dets_subj[0]["box"]
                        o_box = dets_obj[0]["box"]
                        spatial_ok = self._verify_spatial(s_box, o_box, prep, frame)
                        spatial_scores.append(1.0 if spatial_ok else 0.0)
                    else:
                        spatial_scores.append(0.0)
            scores["spatial"] = float(np.mean(spatial_scores)) if spatial_scores else 0.5

        # Numeracy
        counts = self._parse_count(caption)
        if counts:
            count_scores = []
            for expected_num, noun in counts:
                frame_counts = []
                for frame in frames[::max(1, len(frames) // 4)]:
                    dets = self._detect_objects(frame, [noun.rstrip("s")])
                    frame_counts.append(len(dets))
                avg_count = np.mean(frame_counts) if frame_counts else 0
                if expected_num == 0:
                    count_scores.append(1.0 if avg_count == 0 else 0.0)
                else:
                    count_scores.append(max(0.0, 1.0 - abs(avg_count - expected_num) / expected_num))
            scores["numeracy"] = float(np.mean(count_scores)) if count_scores else 0.5

        # Scene composition
        scene_scores = []
        for frame in frames[::max(1, len(frames) // 4)]:
            sim = self._clip_sim(frame, caption)
            scene_scores.append(sim)
        scores["scene"] = float(np.mean(scene_scores)) if scene_scores else 0.5

        # Fill defaults for unparsed dimensions
        for key in ["attribute", "object_rel", "action", "spatial", "numeracy", "scene"]:
            if key not in scores:
                scores[key] = 0.5

        return scores

    def _verify_spatial(self, s_box: list, o_box: list, prep: str, frame: np.ndarray) -> bool:
        """Verify spatial relationship between two bounding boxes."""
        s_cx = (s_box[0] + s_box[2]) / 2
        s_cy = (s_box[1] + s_box[3]) / 2
        o_cx = (o_box[0] + o_box[2]) / 2
        o_cy = (o_box[1] + o_box[3]) / 2

        if prep in ("above", "on top of"):
            return s_cy < o_cy
        elif prep in ("below", "under"):
            return s_cy > o_cy
        elif prep == "left of":
            return s_cx < o_cx
        elif prep == "right of":
            return s_cx > o_cx
        elif prep in ("next to", "beside", "near"):
            dist = np.sqrt((s_cx - o_cx) ** 2 + (s_cy - o_cy) ** 2)
            diag = np.sqrt(frame.shape[0] ** 2 + frame.shape[1] ** 2)
            return dist < diag * 0.3
        elif prep in ("behind", "in front of"):
            # Use depth if available
            if self._depth is not None:
                try:
                    from PIL import Image
                    rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    pil_img = Image.fromarray(rgb)
                    depth_result = self._depth(pil_img)
                    depth_map = np.array(depth_result["depth"])
                    s_depth = depth_map[int(s_cy):int(s_box[3]), int(s_box[0]):int(s_box[2])].mean()
                    o_depth = depth_map[int(o_cy):int(o_box[3]), int(o_box[0]):int(o_box[2])].mean()
                    if prep == "behind":
                        return s_depth > o_depth
                    else:
                        return s_depth < o_depth
                except Exception:
                    pass
            return True  # Can't verify without depth
        return True

    # ------------------------------------------------------------------ #
    # Tier 2: CLIP-only                                                    #
    # ------------------------------------------------------------------ #

    def _compute_clip_only(self, frames: list, caption: str) -> Dict[str, float]:
        scores = {}

        # Attribute binding via CLIP
        attrs = self._parse_attributes(caption)
        if attrs:
            attr_scores = []
            for adj, noun in attrs:
                for frame in frames[::max(1, len(frames) // 4)]:
                    pos_text = f"a {adj} {noun}"
                    neg_text = f"a {noun}"
                    sims = self._clip_sim_batch_text(frame, [pos_text, neg_text])
                    attr_scores.append(sims[0])
            scores["attribute"] = float(np.mean(attr_scores)) if attr_scores else 0.5

        # Object relationship via CLIP
        rels = self._parse_relations(caption)
        if rels:
            rel_scores = []
            for subj, rel, obj in rels:
                for frame in frames[::max(1, len(frames) // 4)]:
                    sim = self._clip_sim(frame, f"{subj} {rel} {obj}")
                    rel_scores.append(sim)
            scores["object_rel"] = float(np.mean(rel_scores)) if rel_scores else 0.5

        # Action binding via CLIP temporal
        actions = self._parse_actions(caption)
        if actions:
            action_scores = []
            for noun, verb in actions:
                for frame in frames:
                    sim = self._clip_sim(frame, f"a {noun} {verb}")
                    action_scores.append(sim)
            scores["action"] = float(np.mean(action_scores)) if action_scores else 0.5

        # Scene via CLIP
        scene_scores = []
        for frame in frames[::max(1, len(frames) // 4)]:
            sim = self._clip_sim(frame, caption)
            scene_scores.append(sim)
        scores["scene"] = float(np.mean(scene_scores)) if scene_scores else 0.5

        # Spatial/numeracy not reliable with CLIP only — skip
        for key in ["attribute", "object_rel", "action", "scene"]:
            if key not in scores:
                scores[key] = 0.5

        return scores

    # ------------------------------------------------------------------ #
    # Tier 3: Heuristic                                                    #
    # ------------------------------------------------------------------ #

    def _compute_heuristic(self, frames: list, caption: str) -> Dict[str, float]:
        scores = {}
        caption_lower = caption.lower()

        # Attribute: check if descriptive adjectives are present
        attrs = self._parse_attributes(caption)
        scores["attribute"] = min(1.0, len(attrs) / 3.0) if attrs else 0.5

        # Relations: check if relation words are present
        rels = self._parse_relations(caption)
        scores["object_rel"] = min(1.0, len(rels) / 2.0) if rels else 0.5

        # Action: estimate motion magnitude per frame
        actions = self._parse_actions(caption)
        if actions:
            grays = [cv2.cvtColor(f, cv2.COLOR_BGR2GRAY) for f in frames]
            motion_mags = []
            for i in range(len(grays) - 1):
                diff = np.abs(grays[i + 1].astype(float) - grays[i].astype(float)).mean()
                motion_mags.append(diff)
            avg_motion = float(np.mean(motion_mags)) if motion_mags else 0
            # Actions imply motion — higher motion = better
            scores["action"] = min(avg_motion / 20.0, 1.0)
        else:
            scores["action"] = 0.5

        # Spatial: basic edge/region analysis
        spatials = self._parse_spatial(caption)
        scores["spatial"] = 0.5

        # Numeracy: edge detection for distinct objects
        counts = self._parse_count(caption)
        if counts:
            # Count contours as proxy for object count
            gray = cv2.cvtColor(frames[len(frames) // 2], cv2.COLOR_BGR2GRAY)
            edges = cv2.Canny(gray, 50, 150)
            contours, _ = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
            # Filter small contours
            h, w = gray.shape
            min_area = h * w * 0.01
            large_contours = [c for c in contours if cv2.contourArea(c) > min_area]
            detected_count = len(large_contours)

            count_scores = []
            for expected_num, _ in counts:
                if expected_num == 0:
                    count_scores.append(1.0 if detected_count == 0 else 0.0)
                elif expected_num > 0:
                    count_scores.append(max(0.0, 1.0 - abs(detected_count - expected_num) / (expected_num + 1)))
            scores["numeracy"] = float(np.mean(count_scores)) if count_scores else 0.5
        else:
            scores["numeracy"] = 0.5

        # Scene: temporal consistency of frames
        if len(frames) >= 3:
            grays = [cv2.cvtColor(f, cv2.COLOR_BGR2GRAY) for f in frames]
            hist_corrs = []
            for i in range(len(grays) - 1):
                h1 = cv2.calcHist([grays[i]], [0], None, [64], [0, 256])
                h2 = cv2.calcHist([grays[i + 1]], [0], None, [64], [0, 256])
                corr = cv2.compareHist(h1, h2, cv2.HISTCMP_CORREL)
                hist_corrs.append(max(corr, 0.0))
            scores["scene"] = float(np.mean(hist_corrs))
        else:
            scores["scene"] = 0.5

        return scores



import logging
import cv2
import numpy as np
from typing import Optional, List, Set, Dict

from ayase.models import Sample, ValidationIssue, ValidationSeverity, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)

class ColorConsistencyModule(PipelineModule):
    name = "color_consistency"
    description = "Verifies color attributes in prompt vs video content"
    default_config = {}

    def __init__(self, config=None):
        super().__init__(config)
        self.color_map = {
            "red": ([0, 50, 50], [10, 255, 255]), # Red wraps around 180
            "green": ([36, 50, 50], [86, 255, 255]),
            "blue": ([100, 50, 50], [130, 255, 255]),
            "yellow": ([20, 50, 50], [35, 255, 255]),
            "orange": ([11, 50, 50], [19, 255, 255]),
            "purple": ([131, 50, 50], [160, 255, 255]),
            "black": ([0, 0, 0], [180, 255, 30]),
            "white": ([0, 0, 200], [180, 30, 255]),
            "gray": ([0, 0, 50], [180, 50, 200])
        }
        # Red has two ranges in HSV
        self.red2 = ([170, 50, 50], [180, 255, 255])

    def process(self, sample: Sample) -> Sample:
        if not sample.caption:
            return sample

        caption_text = sample.caption.text.lower()
        mentioned_colors = [c for c in self.color_map.keys() if c in caption_text]
        
        if not mentioned_colors:
            return sample

        image = self._load_image(sample)
        if image is None:
            return sample

        try:
            hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
            total_pixels = hsv.shape[0] * hsv.shape[1]
            
            missing_colors = []
            color_scores = []
            
            for color in mentioned_colors:
                lower = np.array(self.color_map[color][0], dtype="uint8")
                upper = np.array(self.color_map[color][1], dtype="uint8")
                
                mask = cv2.inRange(hsv, lower, upper)
                
                # Handle Red wrap-around
                if color == "red":
                    lower2 = np.array(self.red2[0], dtype="uint8")
                    upper2 = np.array(self.red2[1], dtype="uint8")
                    mask2 = cv2.inRange(hsv, lower2, upper2)
                    mask = cv2.bitwise_or(mask, mask2)
                
                count = cv2.countNonZero(mask)
                ratio = count / total_pixels
                color_scores.append(min(ratio / 0.02, 1.0))
                
                # If color is explicitly mentioned, we expect at least SOME of it (e.g. > 1% or 0.5%)
                if ratio < 0.005: 
                    missing_colors.append(f"{color} ({ratio*100:.2f}%)")
            
            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            if color_scores:
                sample.quality_metrics.color_score = float(np.mean(color_scores)) * 100.0
            else:
                sample.quality_metrics.color_score = 100.0

            if missing_colors:
                 sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.INFO,
                        message=f"Color mismatch: Caption mentions {', '.join(mentioned_colors)} but found negligible amount of: {', '.join(missing_colors)}",
                        details={"missing_colors": missing_colors}
                    )
                )

        except Exception as e:
            logger.warning(f"Color consistency check failed: {e}")

        return sample

    def _load_image(self, sample: Sample) -> Optional[np.ndarray]:
        try:
            if sample.is_video:
                cap = cv2.VideoCapture(str(sample.path))
                frame_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
                cap.set(cv2.CAP_PROP_POS_FRAMES, frame_count // 2)
                ret, frame = cap.read()
                cap.release()
                return frame if ret else None
            else:
                return cv2.imread(str(sample.path))
        except Exception:
            return None



"""Tests for semantic_segmentation_consistency module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_semantic_segmentation_consistency_basics():
    from ayase.modules.semantic_segmentation_consistency import SemanticSegmentationConsistencyModule
    _test_module_basics(SemanticSegmentationConsistencyModule, "semantic_segmentation_consistency")

def test_semantic_segmentation_consistency_video(video_sample):
    from ayase.modules.semantic_segmentation_consistency import SemanticSegmentationConsistencyModule
    video_sample.quality_metrics = QualityMetrics()
    m = SemanticSegmentationConsistencyModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""LPDist (Log-Power Spectral Distance) module.

Full-reference metric comparing mel-spectrogram power between reference
and degraded audio.  Captures spectral envelope differences — useful for
TTS, voice conversion, and audio coding quality assessment.

Score range: 0.0+ (lower = better, 0 = identical spectra).
  <1.0   excellent
  1-2    good
  2-4    fair
  >4     poor

References:
    - Inspired by log-spectral distance measures (Gray & Markel, 1976),
      adapted to mel-spectrogram domain for modern audio quality assessment
"""

import logging
from pathlib import Path
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample, ValidationIssue, ValidationSeverity
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AudioLPDistModule(PipelineModule):
    name = "audio_lpdist"
    description = "Log-Power Spectral Distance (full-reference audio)"
    default_config = {
        "target_sr": 16000,
        "n_mels": 80,
        "warning_threshold": 4.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.target_sr = self.config.get("target_sr", 16000)
        self.n_mels = self.config.get("n_mels", 80)
        self.warning_threshold = self.config.get("warning_threshold", 4.0)
        self._ml_available = False

    def setup(self) -> None:
        try:
            import librosa  # noqa: F401

            self._ml_available = True
            logger.info("LPDist module initialised (librosa)")
        except ImportError:
            logger.warning("librosa not installed. Install with: pip install librosa")

    def process(self, sample: Sample) -> Sample:
        if not self._ml_available:
            return sample

        reference = getattr(sample, "reference_path", None)
        if reference is None:
            return sample
        reference = Path(reference) if not isinstance(reference, Path) else reference
        if not reference.exists():
            return sample

        try:
            ref_mel = self._extract_log_mel(reference)
            deg_mel = self._extract_log_mel(sample.path)

            if ref_mel is None or deg_mel is None:
                return sample

            min_frames = min(ref_mel.shape[1], deg_mel.shape[1])
            if min_frames < 5:
                return sample

            ref_mel = ref_mel[:, :min_frames]
            deg_mel = deg_mel[:, :min_frames]

            # Mean squared difference in log-power domain
            lpdist = float(np.sqrt(np.mean((ref_mel - deg_mel) ** 2)))

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.lpdist_score = round(lpdist, 4)

            if lpdist > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"High spectral distance (LPDist): {lpdist:.3f}",
                        details={"lpdist": lpdist},
                    )
                )

        except Exception as e:
            logger.warning(f"LPDist failed for {sample.path}: {e}")

        return sample

    def _extract_log_mel(self, path: Path) -> Optional[np.ndarray]:
        try:
            import librosa

            y, sr = librosa.load(str(path), sr=self.target_sr, mono=True)
            mel = librosa.feature.melspectrogram(
                y=y, sr=sr, n_mels=self.n_mels
            )
            log_mel = librosa.power_to_db(mel, ref=np.max)
            return log_mel
        except Exception as e:
            logger.debug(f"Mel extraction failed: {e}")
            return None



"""Tests for xpsnr module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_xpsnr_basics():
    from ayase.modules.xpsnr import XPSNRModule
    _test_module_basics(XPSNRModule, "xpsnr")

def test_xpsnr_no_reference(image_sample):
    from ayase.modules.xpsnr import XPSNRModule
    m = XPSNRModule()
    result = m.process(image_sample)
    assert result is image_sample



# ------------------------------------------------------------------------
# Deformable DETR
# Copyright (c) 2020 SenseTime. All Rights Reserved.
# Licensed under the Apache License, Version 2.0 [see LICENSE for details]
# ------------------------------------------------------------------------
# Modified from DETR (https://github.com/facebookresearch/detr)
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
# ------------------------------------------------------------------------

"""
This file provides the definition of the convolutional heads used to predict masks, as well as the losses
"""
import io
from collections import defaultdict

import torch
import torch.nn as nn
import torch.nn.functional as F
from PIL import Image

import util.box_ops as box_ops
from util.misc import NestedTensor, interpolate, nested_tensor_from_tensor_list

try:
    from panopticapi.utils import id2rgb, rgb2id
except ImportError:
    pass


class DETRsegm(nn.Module):
    def __init__(self, detr, freeze_detr=False):
        super().__init__()
        self.detr = detr

        if freeze_detr:
            for p in self.parameters():
                p.requires_grad_(False)

        hidden_dim, nheads = detr.transformer.d_model, detr.transformer.nhead
        self.bbox_attention = MHAttentionMap(hidden_dim, hidden_dim, nheads, dropout=0)
        self.mask_head = MaskHeadSmallConv(hidden_dim + nheads, [1024, 512, 256], hidden_dim)

    def forward(self, samples: NestedTensor):
        if not isinstance(samples, NestedTensor):
            samples = nested_tensor_from_tensor_list(samples)
        features, pos = self.detr.backbone(samples)

        bs = features[-1].tensors.shape[0]

        src, mask = features[-1].decompose()
        src_proj = self.detr.input_proj(src)
        hs, memory = self.detr.transformer(src_proj, mask, self.detr.query_embed.weight, pos[-1])

        outputs_class = self.detr.class_embed(hs)
        outputs_coord = self.detr.bbox_embed(hs).sigmoid()
        out = {"pred_logits": outputs_class[-1], "pred_boxes": outputs_coord[-1]}
        if self.detr.aux_loss:
            out["aux_outputs"] = [
                {"pred_logits": a, "pred_boxes": b} for a, b in zip(outputs_class[:-1], outputs_coord[:-1])
            ]

        # FIXME h_boxes takes the last one computed, keep this in mind
        bbox_mask = self.bbox_attention(hs[-1], memory, mask=mask)

        seg_masks = self.mask_head(src_proj, bbox_mask, [features[2].tensors, features[1].tensors, features[0].tensors])
        outputs_seg_masks = seg_masks.view(bs, self.detr.num_queries, seg_masks.shape[-2], seg_masks.shape[-1])

        out["pred_masks"] = outputs_seg_masks
        return out


class MaskHeadSmallConv(nn.Module):
    """
    Simple convolutional head, using group norm.
    Upsampling is done using a FPN approach
    """

    def __init__(self, dim, fpn_dims, context_dim):
        super().__init__()

        inter_dims = [dim, context_dim // 2, context_dim // 4, context_dim // 8, context_dim // 16, context_dim // 64]
        self.lay1 = torch.nn.Conv2d(dim, dim, 3, padding=1)
        self.gn1 = torch.nn.GroupNorm(8, dim)
        self.lay2 = torch.nn.Conv2d(dim, inter_dims[1], 3, padding=1)
        self.gn2 = torch.nn.GroupNorm(8, inter_dims[1])
        self.lay3 = torch.nn.Conv2d(inter_dims[1], inter_dims[2], 3, padding=1)
        self.gn3 = torch.nn.GroupNorm(8, inter_dims[2])
        self.lay4 = torch.nn.Conv2d(inter_dims[2], inter_dims[3], 3, padding=1)
        self.gn4 = torch.nn.GroupNorm(8, inter_dims[3])
        self.lay5 = torch.nn.Conv2d(inter_dims[3], inter_dims[4], 3, padding=1)
        self.gn5 = torch.nn.GroupNorm(8, inter_dims[4])
        self.out_lay = torch.nn.Conv2d(inter_dims[4], 1, 3, padding=1)

        self.dim = dim

        self.adapter1 = torch.nn.Conv2d(fpn_dims[0], inter_dims[1], 1)
        self.adapter2 = torch.nn.Conv2d(fpn_dims[1], inter_dims[2], 1)
        self.adapter3 = torch.nn.Conv2d(fpn_dims[2], inter_dims[3], 1)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight, a=1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x, bbox_mask, fpns):
        def expand(tensor, length):
            return tensor.unsqueeze(1).repeat(1, int(length), 1, 1, 1).flatten(0, 1)

        x = torch.cat([expand(x, bbox_mask.shape[1]), bbox_mask.flatten(0, 1)], 1)

        x = self.lay1(x)
        x = self.gn1(x)
        x = F.relu(x)
        x = self.lay2(x)
        x = self.gn2(x)
        x = F.relu(x)

        cur_fpn = self.adapter1(fpns[0])
        if cur_fpn.size(0) != x.size(0):
            cur_fpn = expand(cur_fpn, x.size(0) / cur_fpn.size(0))
        x = cur_fpn + F.interpolate(x, size=cur_fpn.shape[-2:], mode="nearest")
        x = self.lay3(x)
        x = self.gn3(x)
        x = F.relu(x)

        cur_fpn = self.adapter2(fpns[1])
        if cur_fpn.size(0) != x.size(0):
            cur_fpn = expand(cur_fpn, x.size(0) / cur_fpn.size(0))
        x = cur_fpn + F.interpolate(x, size=cur_fpn.shape[-2:], mode="nearest")
        x = self.lay4(x)
        x = self.gn4(x)
        x = F.relu(x)

        cur_fpn = self.adapter3(fpns[2])
        if cur_fpn.size(0) != x.size(0):
            cur_fpn = expand(cur_fpn, x.size(0) / cur_fpn.size(0))
        x = cur_fpn + F.interpolate(x, size=cur_fpn.shape[-2:], mode="nearest")
        x = self.lay5(x)
        x = self.gn5(x)
        x = F.relu(x)

        x = self.out_lay(x)
        return x


class MHAttentionMap(nn.Module):
    """This is a 2D attention module, which only returns the attention softmax (no multiplication by value)"""

    def __init__(self, query_dim, hidden_dim, num_heads, dropout=0, bias=True):
        super().__init__()
        self.num_heads = num_heads
        self.hidden_dim = hidden_dim
        self.dropout = nn.Dropout(dropout)

        self.q_linear = nn.Linear(query_dim, hidden_dim, bias=bias)
        self.k_linear = nn.Linear(query_dim, hidden_dim, bias=bias)

        nn.init.zeros_(self.k_linear.bias)
        nn.init.zeros_(self.q_linear.bias)
        nn.init.xavier_uniform_(self.k_linear.weight)
        nn.init.xavier_uniform_(self.q_linear.weight)
        self.normalize_fact = float(hidden_dim / self.num_heads) ** -0.5

    def forward(self, q, k, mask=None):
        q = self.q_linear(q)
        k = F.conv2d(k, self.k_linear.weight.unsqueeze(-1).unsqueeze(-1), self.k_linear.bias)
        qh = q.view(q.shape[0], q.shape[1], self.num_heads, self.hidden_dim // self.num_heads)
        kh = k.view(k.shape[0], self.num_heads, self.hidden_dim // self.num_heads, k.shape[-2], k.shape[-1])
        weights = torch.einsum("bqnc,bnchw->bqnhw", qh * self.normalize_fact, kh)

        if mask is not None:
            weights.masked_fill_(mask.unsqueeze(1).unsqueeze(1), float("-inf"))
        weights = F.softmax(weights.flatten(2), dim=-1).view_as(weights)
        weights = self.dropout(weights)
        return weights


def dice_loss(inputs, targets, num_boxes):
    """
    Compute the DICE loss, similar to generalized IOU for masks
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
    """
    inputs = inputs.sigmoid()
    inputs = inputs.flatten(1)
    numerator = 2 * (inputs * targets).sum(1)
    denominator = inputs.sum(-1) + targets.sum(-1)
    loss = 1 - (numerator + 1) / (denominator + 1)
    return loss.sum() / num_boxes


def sigmoid_focal_loss(inputs, targets, num_boxes, alpha: float = 0.25, gamma: float = 2):
    """
    Loss used in RetinaNet for dense detection: https://arxiv.org/abs/1708.02002.
    Args:
        inputs: A float tensor of arbitrary shape.
                The predictions for each example.
        targets: A float tensor with the same shape as inputs. Stores the binary
                 classification label for each element in inputs
                (0 for the negative class and 1 for the positive class).
        alpha: (optional) Weighting factor in range (0,1) to balance
                positive vs negative examples. Default = -1 (no weighting).
        gamma: Exponent of the modulating factor (1 - p_t) to
               balance easy vs hard examples.
    Returns:
        Loss tensor
    """
    prob = inputs.sigmoid()
    ce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")
    p_t = prob * targets + (1 - prob) * (1 - targets)
    loss = ce_loss * ((1 - p_t) ** gamma)

    if alpha >= 0:
        alpha_t = alpha * targets + (1 - alpha) * (1 - targets)
        loss = alpha_t * loss

    return loss.mean(1).sum() / num_boxes


class PostProcessSegm(nn.Module):
    def __init__(self, threshold=0.5):
        super().__init__()
        self.threshold = threshold

    @torch.no_grad()
    def forward(self, results, outputs, orig_target_sizes, max_target_sizes):
        assert len(orig_target_sizes) == len(max_target_sizes)
        max_h, max_w = max_target_sizes.max(0)[0].tolist()
        outputs_masks = outputs["pred_masks"].squeeze(2)
        outputs_masks = F.interpolate(outputs_masks, size=(max_h, max_w), mode="bilinear", align_corners=False)
        outputs_masks = (outputs_masks.sigmoid() > self.threshold).cpu()

        for i, (cur_mask, t, tt) in enumerate(zip(outputs_masks, max_target_sizes, orig_target_sizes)):
            img_h, img_w = t[0], t[1]
            results[i]["masks"] = cur_mask[:, :img_h, :img_w].unsqueeze(1)
            results[i]["masks"] = F.interpolate(
                results[i]["masks"].float(), size=tuple(tt.tolist()), mode="nearest"
            ).byte()

        return results


class PostProcessPanoptic(nn.Module):
    """This class converts the output of the model to the final panoptic result, in the format expected by the
    coco panoptic API """

    def __init__(self, is_thing_map, threshold=0.85):
        """
        Parameters:
           is_thing_map: This is a whose keys are the class ids, and the values a boolean indicating whether
                          the class is  a thing (True) or a stuff (False) class
           threshold: confidence threshold: segments with confidence lower than this will be deleted
        """
        super().__init__()
        self.threshold = threshold
        self.is_thing_map = is_thing_map

    def forward(self, outputs, processed_sizes, target_sizes=None):
        """ This function computes the panoptic prediction from the model's predictions.
        Parameters:
            outputs: This is a dict coming directly from the model. See the model doc for the content.
            processed_sizes: This is a list of tuples (or torch tensors) of sizes of the images that were passed to the
                             model, ie the size after data augmentation but before batching.
            target_sizes: This is a list of tuples (or torch tensors) corresponding to the requested final size
                          of each prediction. If left to None, it will default to the processed_sizes
            """
        if target_sizes is None:
            target_sizes = processed_sizes
        assert len(processed_sizes) == len(target_sizes)
        out_logits, raw_masks, raw_boxes = outputs["pred_logits"], outputs["pred_masks"], outputs["pred_boxes"]
        assert len(out_logits) == len(raw_masks) == len(target_sizes)
        preds = []

        def to_tuple(tup):
            if isinstance(tup, tuple):
                return tup
            return tuple(tup.cpu().tolist())

        for cur_logits, cur_masks, cur_boxes, size, target_size in zip(
            out_logits, raw_masks, raw_boxes, processed_sizes, target_sizes
        ):
            # we filter empty queries and detection below threshold
            scores, labels = cur_logits.softmax(-1).max(-1)
            keep = labels.ne(outputs["pred_logits"].shape[-1] - 1) & (scores > self.threshold)
            cur_scores, cur_classes = cur_logits.softmax(-1).max(-1)
            cur_scores = cur_scores[keep]
            cur_classes = cur_classes[keep]
            cur_masks = cur_masks[keep]
            cur_masks = interpolate(cur_masks[None], to_tuple(size), mode="bilinear").squeeze(0)
            cur_boxes = box_ops.box_cxcywh_to_xyxy(cur_boxes[keep])

            h, w = cur_masks.shape[-2:]
            assert len(cur_boxes) == len(cur_classes)

            # It may be that we have several predicted masks for the same stuff class.
            # In the following, we track the list of masks ids for each stuff class (they are merged later on)
            cur_masks = cur_masks.flatten(1)
            stuff_equiv_classes = defaultdict(lambda: [])
            for k, label in enumerate(cur_classes):
                if not self.is_thing_map[label.item()]:
                    stuff_equiv_classes[label.item()].append(k)

            def get_ids_area(masks, scores, dedup=False):
                # This helper function creates the final panoptic segmentation image
                # It also returns the area of the masks that appears on the image

                m_id = masks.transpose(0, 1).softmax(-1)

                if m_id.shape[-1] == 0:
                    # We didn't detect any mask :(
                    m_id = torch.zeros((h, w), dtype=torch.long, device=m_id.device)
                else:
                    m_id = m_id.argmax(-1).view(h, w)

                if dedup:
                    # Merge the masks corresponding to the same stuff class
                    for equiv in stuff_equiv_classes.values():
                        if len(equiv) > 1:
                            for eq_id in equiv:
                                m_id.masked_fill_(m_id.eq(eq_id), equiv[0])

                final_h, final_w = to_tuple(target_size)

                seg_img = Image.fromarray(id2rgb(m_id.view(h, w).cpu().numpy()))
                seg_img = seg_img.resize(size=(final_w, final_h), resample=Image.NEAREST)

                np_seg_img = (
                    torch.ByteTensor(torch.ByteStorage.from_buffer(seg_img.tobytes())).view(final_h, final_w, 3).numpy()
                )
                m_id = torch.from_numpy(rgb2id(np_seg_img))

                area = []
                for i in range(len(scores)):
                    area.append(m_id.eq(i).sum().item())
                return area, seg_img

            area, seg_img = get_ids_area(cur_masks, cur_scores, dedup=True)
            if cur_classes.numel() > 0:
                # We know filter empty masks as long as we find some
                while True:
                    filtered_small = torch.as_tensor(
                        [area[i] <= 4 for i, c in enumerate(cur_classes)], dtype=torch.bool, device=keep.device
                    )
                    if filtered_small.any().item():
                        cur_scores = cur_scores[~filtered_small]
                        cur_classes = cur_classes[~filtered_small]
                        cur_masks = cur_masks[~filtered_small]
                        area, seg_img = get_ids_area(cur_masks, cur_scores)
                    else:
                        break

            else:
                cur_classes = torch.ones(1, dtype=torch.long, device=cur_classes.device)

            segments_info = []
            for i, a in enumerate(area):
                cat = cur_classes[i].item()
                segments_info.append({"id": i, "isthing": self.is_thing_map[cat], "category_id": cat, "area": a})
            del cur_classes

            with io.BytesIO() as out:
                seg_img.save(out, format="PNG")
                predictions = {"png_string": out.getvalue(), "segments_info": segments_info}
            preds.append(predictions)
        return preds



"""Tests for t2veval module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_t2veval_basics():
    from ayase.modules.t2veval import T2VEvalModule
    _test_module_basics(T2VEvalModule, "t2veval")

def test_t2veval_image(image_sample):
    from ayase.modules.t2veval import T2VEvalModule
    image_sample.quality_metrics = QualityMetrics()
    m = T2VEvalModule()
    m.on_mount()
    result = m.process(image_sample)
    assert result is image_sample

def test_t2veval_video(video_sample):
    from ayase.modules.t2veval import T2VEvalModule
    video_sample.quality_metrics = QualityMetrics()
    m = T2VEvalModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



import logging
import cv2
import numpy as np
from typing import List

from ayase.models import Sample, ValidationIssue, ValidationSeverity, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class TemporalFlickeringModule(PipelineModule):
    name = "temporal_flickering"
    description = "Warping Error using RAFT optical flow with occlusion masking"

    default_config = {
        "warning_threshold": 0.02,
        "max_frames": 300,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.warning_threshold = self.config.get("warning_threshold", 0.02)
        self._model = None
        self._device = "cpu"
        self._ml_available = False
        self._transforms = None

    def setup(self) -> None:
        try:
            import os
            import torch
            from torchvision.models.optical_flow import raft_small, Raft_Small_Weights

            # Redirect torch hub cache to models_dir so RAFT weights respect config
            models_dir = self.config.get("models_dir")
            if models_dir:
                os.environ["TORCH_HOME"] = str(models_dir)

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            logger.info(f"Setting up RAFT-Small for warping error on {self._device}...")

            weights = Raft_Small_Weights.DEFAULT
            self._model = raft_small(weights=weights, progress=False).to(self._device)
            self._model.eval()
            self._transforms = weights.transforms()
            self._ml_available = True

        except ImportError:
            logger.warning("torchvision >= 0.13 required for RAFT.")
        except Exception as e:
            logger.warning(f"Failed to setup RAFT: {e}")

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        if self._ml_available:
            self._analyze_raft(sample)
        else:
            self._analyze_farneback_fallback(sample)

        return sample

    def _analyze_raft(self, sample: Sample) -> None:
        """RAFT-based warping error (matches EvalCrafter implementation).

        For each consecutive pair:
        1. Compute forward and backward flow with RAFT (downsampled 2x)
        2. Detect occlusions via forward-backward consistency
        3. Warp frame2 to frame1 using forward flow
        4. MSE in non-occluded regions
        """
        import torch
        import torch.nn.functional as F

        try:
            frames = self._load_all_frames(sample)
            if len(frames) < 2:
                return

            total_err = 0.0

            with torch.no_grad():
                for i in range(len(frames) - 1):
                    # Convert to tensor [1, C, H, W] float [0, 255]
                    img1 = torch.from_numpy(frames[i]).permute(2, 0, 1).float().unsqueeze(0).to(self._device) / 255.0
                    img2 = torch.from_numpy(frames[i + 1]).permute(2, 0, 1).float().unsqueeze(0).to(self._device) / 255.0

                    # Downsample by 2x (matches EvalCrafter)
                    img1 = F.interpolate(img1, scale_factor=0.5, mode="bilinear", align_corners=False)
                    img2 = F.interpolate(img2, scale_factor=0.5, mode="bilinear", align_corners=False)

                    # Pad to multiple of 8 (RAFT requirement)
                    _, _, h, w = img1.shape
                    pad_h = (8 - h % 8) % 8
                    pad_w = (8 - w % 8) % 8
                    if pad_h > 0 or pad_w > 0:
                        img1 = F.pad(img1, (0, pad_w, 0, pad_h), mode="reflect")
                        img2 = F.pad(img2, (0, pad_w, 0, pad_h), mode="reflect")

                    # Prepare for RAFT (needs [0, 255] range after transforms)
                    img1_raft = img1 * 255.0
                    img2_raft = img2 * 255.0

                    if self._transforms:
                        img1_t, img2_t = self._transforms(img1_raft, img2_raft)
                    else:
                        img1_t, img2_t = img1_raft, img2_raft

                    # Forward flow (frame1 → frame2)
                    fw_flow = self._model(img1_t, img2_t)[-1]
                    # Backward flow (frame2 → frame1)
                    bw_flow = self._model(img2_t, img1_t)[-1]

                    # Crop back to original size (remove padding)
                    if pad_h > 0 or pad_w > 0:
                        crop_h = fw_flow.shape[2] - pad_h if pad_h > 0 else fw_flow.shape[2]
                        crop_w = fw_flow.shape[3] - pad_w if pad_w > 0 else fw_flow.shape[3]
                        fw_flow = fw_flow[:, :, :crop_h, :crop_w]
                        bw_flow = bw_flow[:, :, :crop_h, :crop_w]
                        img1 = img1[:, :, :crop_h, :crop_w]
                        img2 = img2[:, :, :crop_h, :crop_w]

                    # Warp img2 to img1 using forward flow
                    warped_img2 = self._warp(img2, fw_flow)

                    # Detect occlusions via forward-backward consistency
                    occ_mask = self._detect_occlusion(fw_flow, bw_flow)
                    noc_mask = 1.0 - occ_mask

                    # MSE in non-occluded regions
                    diff = (warped_img2 - img1) * noc_mask
                    diff_sq = diff ** 2

                    n_pixels = torch.sum(noc_mask)
                    if n_pixels == 0:
                        n_pixels = torch.tensor(diff_sq.numel(), dtype=torch.float32)

                    total_err += (torch.sum(diff_sq) / n_pixels).item()

            warping_error = total_err / (len(frames) - 1)

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()
            sample.quality_metrics.warping_error = float(warping_error)

            if warping_error > self.warning_threshold:
                sample.validation_issues.append(
                    ValidationIssue(
                        severity=ValidationSeverity.WARNING,
                        message=f"High flickering detected (Warping Error): {warping_error:.4f}",
                        details={"warping_error": float(warping_error)},
                    )
                )

        except Exception as e:
            logger.warning(f"RAFT warping error failed: {e}")
            self._analyze_farneback_fallback(sample)

    def _warp(self, img, flow):
        """Warp image using optical flow via grid_sample."""
        import torch
        import torch.nn.functional as F

        _, _, h, w = img.shape
        # Create normalized grid
        grid_y, grid_x = torch.meshgrid(
            torch.arange(h, device=flow.device, dtype=torch.float32),
            torch.arange(w, device=flow.device, dtype=torch.float32),
            indexing="ij",
        )
        # Add flow displacement
        grid_x = grid_x + flow[0, 0]
        grid_y = grid_y + flow[0, 1]
        # Normalize to [-1, 1]
        grid_x = 2.0 * grid_x / (w - 1) - 1.0
        grid_y = 2.0 * grid_y / (h - 1) - 1.0
        grid = torch.stack([grid_x, grid_y], dim=-1).unsqueeze(0)

        return F.grid_sample(img, grid, mode="bilinear", padding_mode="zeros", align_corners=True)

    def _detect_occlusion(self, fw_flow, bw_flow, threshold=1.0):
        """Detect occlusions via forward-backward flow consistency."""
        import torch

        # Warp backward flow using forward flow
        _, _, h, w = fw_flow.shape
        grid_y, grid_x = torch.meshgrid(
            torch.arange(h, device=fw_flow.device, dtype=torch.float32),
            torch.arange(w, device=fw_flow.device, dtype=torch.float32),
            indexing="ij",
        )
        map_x = grid_x + fw_flow[0, 0]
        map_y = grid_y + fw_flow[0, 1]
        # Normalize
        map_x = 2.0 * map_x / (w - 1) - 1.0
        map_y = 2.0 * map_y / (h - 1) - 1.0
        grid = torch.stack([map_x, map_y], dim=-1).unsqueeze(0)

        import torch.nn.functional as F
        warped_bw = F.grid_sample(bw_flow, grid, mode="bilinear", padding_mode="zeros", align_corners=True)

        # Check consistency: ||fw + warped_bw|| should be small for non-occluded
        consistency = fw_flow + warped_bw
        mag = torch.sqrt(consistency[:, 0] ** 2 + consistency[:, 1] ** 2)
        occ = (mag > threshold).float().unsqueeze(1)

        return occ

    def _analyze_farneback_fallback(self, sample: Sample) -> None:
        """Fallback: Farneback-based warping error when RAFT is unavailable."""
        cap = cv2.VideoCapture(str(sample.path))
        if not cap.isOpened():
            return

        prev_gray = None
        warping_errors = []
        frame_idx = 0

        while True:
            ret, frame = cap.read()
            if not ret:
                break

            gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

            if prev_gray is not None:
                flow = cv2.calcOpticalFlowFarneback(prev_gray, gray, None, 0.5, 3, 15, 3, 5, 1.2, 0)
                h, w = gray.shape
                grid_x, grid_y = np.meshgrid(np.arange(w), np.arange(h))
                map_x = (grid_x + flow[..., 0]).astype(np.float32)
                map_y = (grid_y + flow[..., 1]).astype(np.float32)
                warped_prev = cv2.remap(prev_gray, map_x, map_y, cv2.INTER_LINEAR)

                diff = (gray.astype(np.float32) / 255.0 - warped_prev.astype(np.float32) / 255.0) ** 2
                warping_errors.append(np.mean(diff))

            prev_gray = gray
            frame_idx += 1

        cap.release()

        if not warping_errors:
            return

        avg_error = float(np.mean(warping_errors))
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        sample.quality_metrics.warping_error = avg_error

        if avg_error > self.warning_threshold:
            sample.validation_issues.append(
                ValidationIssue(
                    severity=ValidationSeverity.WARNING,
                    message=f"High flickering detected (Warping Error, Farneback fallback): {avg_error:.4f}",
                    details={"warping_error": avg_error},
                )
            )

    def _load_all_frames(self, sample: Sample) -> List[np.ndarray]:
        max_frames = self.config.get("max_frames", 300)
        frames = []
        cap = None
        try:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

            if total > max_frames and total > 0:
                # Subsample uniformly to stay within memory budget
                indices = set(np.linspace(0, total - 1, max_frames, dtype=int))
                idx = 0
                while cap.isOpened():
                    ret, frame = cap.read()
                    if not ret:
                        break
                    if idx in indices:
                        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                        frames.append(frame)
                    idx += 1
            else:
                while cap.isOpened():
                    ret, frame = cap.read()
                    if not ret:
                        break
                    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    frames.append(frame)
        except Exception as e:
            logger.debug(f"Failed to load frames: {e}")
        finally:
            if cap is not None:
                cap.release()
        return frames



"""ChronoMagic-Bench module — NeurIPS 2024 (arXiv:2406.18522).

Two metrics for time-lapse video quality:
  - **MTScore** (Metamorphic Temporal): smoothness of temporal progression
  - **CHScore** (Chrono-Hallucination): detection of temporal hallucinations

Backend tiers:
  1. **CLIP** — CLIP ViT-B/32 temporal embedding analysis
  2. **Heuristic** — Pixel-level temporal gradient + histogram correlation

Video-only: returns None for images.
"""

import logging
from typing import Optional, Tuple

import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class ChronoMagicModule(PipelineModule):
    name = "chronomagic"
    description = "ChronoMagic-Bench MTScore + CHScore (CLIP / heuristic)"
    default_config = {
        "subsample": 16,
        "hallucination_threshold": 2.0,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._backend = "heuristic"
        self._clip_model = None
        self._clip_processor = None
        self._device = "cpu"

    def setup(self) -> None:
        # Tier 1: CLIP
        try:
            import torch
            from transformers import CLIPModel, CLIPProcessor

            self._device = "cuda" if torch.cuda.is_available() else "cpu"
            models_dir = self.config.get("models_dir", "models")

            self._clip_model = CLIPModel.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir=models_dir,
            ).to(self._device)
            self._clip_model.eval()
            self._clip_processor = CLIPProcessor.from_pretrained(
                "openai/clip-vit-base-patch32", cache_dir=models_dir,
            )
            self._backend = "clip"
            logger.info("ChronoMagic loaded CLIP on %s", self._device)
            return
        except Exception as e:
            logger.info("CLIP unavailable for ChronoMagic: %s", e)

        logger.info("ChronoMagic using heuristic backend")

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        try:
            if self._backend == "clip":
                mt_score, ch_score = self._compute_clip(sample)
            else:
                mt_score, ch_score = self._compute_heuristic(sample)

            if mt_score is not None:
                sample.quality_metrics.chronomagic_mt_score = mt_score
            if ch_score is not None:
                sample.quality_metrics.chronomagic_ch_score = ch_score

        except Exception as e:
            logger.warning("ChronoMagic processing failed: %s", e)

        return sample

    # ------------------------------------------------------------------ #
    # Shared frame extraction                                              #
    # ------------------------------------------------------------------ #

    def _extract_frames(self, sample: Sample) -> list:
        num_frames = self.config.get("subsample", 16)
        cap = cv2.VideoCapture(str(sample.path))
        total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        if total < 4:
            cap.release()
            return []

        indices = list(range(0, total, max(1, total // num_frames)))[:num_frames]
        frames = []
        for idx in indices:
            cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
            ret, frame = cap.read()
            if ret:
                frames.append(frame)
        cap.release()
        return frames

    # ------------------------------------------------------------------ #
    # Tier 1: CLIP temporal analysis                                       #
    # ------------------------------------------------------------------ #

    def _compute_clip(self, sample: Sample) -> Tuple[Optional[float], Optional[float]]:
        import torch
        from PIL import Image

        frames = self._extract_frames(sample)
        if len(frames) < 4:
            return None, None

        # Get CLIP embeddings for each frame
        embeddings = []
        for frame in frames:
            rgb = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            pil_img = Image.fromarray(rgb)
            inputs = self._clip_processor(images=pil_img, return_tensors="pt").to(self._device)
            with torch.no_grad():
                features = self._clip_model.get_image_features(**inputs)
                features = features / features.norm(dim=-1, keepdim=True)
                embeddings.append(features.cpu().numpy().flatten())

        embeddings = np.array(embeddings)  # [T, D]

        # MTScore: temporal gradient smoothness
        gradients = np.diff(embeddings, axis=0)  # [T-1, D]
        gradient_norms = np.linalg.norm(gradients, axis=1)  # [T-1]

        # Smoothness = low variance of gradient norms (consistent rate of change)
        if gradient_norms.std() > 1e-8:
            gradient_cv = gradient_norms.std() / (gradient_norms.mean() + 1e-8)
            gradient_smoothness = 1.0 / (1.0 + gradient_cv)
        else:
            gradient_smoothness = 1.0

        # Progression quality: monotonic progression in embedding space
        cosine_sims = []
        for i in range(len(embeddings) - 1):
            sim = float(np.dot(embeddings[i], embeddings[i + 1]))
            cosine_sims.append(sim)
        cosine_sims = np.array(cosine_sims)

        # Good progression = consistently high but not perfect similarity
        progression_quality = float(np.mean(cosine_sims))
        progression_quality = min(max(progression_quality, 0.0), 1.0)

        mt_score = 0.5 * gradient_smoothness + 0.5 * progression_quality

        # CHScore: detect abrupt embedding spikes (hallucination frames)
        threshold = self.config.get("hallucination_threshold", 2.0)
        mean_norm = gradient_norms.mean()
        std_norm = gradient_norms.std() + 1e-8
        hallucination_mask = gradient_norms > (mean_norm + threshold * std_norm)
        hallucination_frac = float(np.mean(hallucination_mask))
        ch_score = 1.0 - hallucination_frac  # Higher = fewer hallucinations

        return (
            float(np.clip(mt_score, 0.0, 1.0)),
            float(np.clip(ch_score, 0.0, 1.0)),
        )

    # ------------------------------------------------------------------ #
    # Tier 2: Heuristic (pixel-level)                                      #
    # ------------------------------------------------------------------ #

    def _compute_heuristic(self, sample: Sample) -> Tuple[Optional[float], Optional[float]]:
        frames = self._extract_frames(sample)
        if len(frames) < 4:
            return None, None

        # Convert to grayscale
        grays = [cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float32) for f in frames]

        # Temporal pixel differences
        diffs = []
        for i in range(len(grays) - 1):
            diff = np.abs(grays[i + 1] - grays[i]).mean()
            diffs.append(diff)
        diffs = np.array(diffs)

        # MTScore: smoothness of temporal gradients
        if diffs.std() > 1e-6:
            cv = diffs.std() / (diffs.mean() + 1e-6)
            gradient_smoothness = 1.0 / (1.0 + cv)
        else:
            gradient_smoothness = 1.0

        # Histogram correlation for progression quality
        hist_sims = []
        for i in range(len(grays) - 1):
            h1 = cv2.calcHist([grays[i].astype(np.uint8)], [0], None, [64], [0, 256])
            h2 = cv2.calcHist([grays[i + 1].astype(np.uint8)], [0], None, [64], [0, 256])
            corr = cv2.compareHist(h1, h2, cv2.HISTCMP_CORREL)
            hist_sims.append(max(corr, 0.0))
        progression_quality = float(np.mean(hist_sims)) if hist_sims else 0.5

        mt_score = 0.5 * gradient_smoothness + 0.5 * progression_quality

        # CHScore: detect abrupt changes (hallucination)
        threshold = self.config.get("hallucination_threshold", 2.0)
        mean_diff = diffs.mean()
        std_diff = diffs.std() + 1e-6
        hallucination_mask = diffs > (mean_diff + threshold * std_diff)
        ch_score = 1.0 - float(np.mean(hallucination_mask))

        return (
            float(np.clip(mt_score, 0.0, 1.0)),
            float(np.clip(ch_score, 0.0, 1.0)),
        )



import logging
from contextlib import suppress

import torch
import torch.nn.functional as F
from tqdm import tqdm

# from open_clip import image_to_device


def evaluate(
    model,
    dataloader,
    tokenizer,
    device,
    video_dataset=False,
    amp=True,
    recall_k_list=[1, 5, 10],
    args=None,
):
    """
    Evaluate the model on the given dataset

    Parameters
    ----------

    model: torch.nn,Module
        CLIP-like model with `encode_image` and `encode_text`

    dataloader: torch.utils.data.Dataloader
        dataloader to use for evaluation

    tokenizer:
        text tokenizer, i.e. convert list of strings to torch.Tensor of integers

    device: cpu/cuda

    amp: whether to use automatic mixed precision

    recall_k_list: list of int
        recall@k k's to use

    Returns
    -------

    dict of retrieval metrics
    """
    # list of batch of images embedding
    batch_images_emb_list = []
    # list of batch of text embedding
    batch_texts_emb_list = []
    # for each text, we collect the corresponding image index, as each image can have multiple corresponding texts
    texts_image_index = []
    dataloader_wrapper = dataloader_with_indices(dataloader)
    autocast = torch.cuda.amp.autocast if amp else suppress
    for batch_images, batch_texts, inds in tqdm(dataloader_wrapper):
        # move the batch to the device

        if isinstance(batch_images, torch.Tensor):
            batch_images = batch_images.to(device, torch.float32)
        elif isinstance(batch_images, (list, tuple)):  # video frames as a list/tuple
            batch_images = [x.to(device, torch.float32) for x in batch_images]
            batch_images = torch.stack(batch_images, dim=0).permute(1,0,2,3,4).contiguous() # nbchw -> bncwh            
        else:
            raise NotImplementedError

        # tokenize all texts in the batch
        batch_texts_tok = tokenizer(
            [text for i, texts in enumerate(batch_texts) for text in texts]
        ).to(device)

        # store the index of image for each text
        if isinstance(inds, torch.Tensor):
            inds = inds.tolist()
        batch_texts_image_index = [
            ind for ind, texts in zip(inds, batch_texts) for text in texts
        ]
        # compute the embedding of images and texts
        with torch.no_grad(), autocast():
            if video_dataset:
                image_features = model.encode_video(batch_images)
            else:
                image_features = model.encode_image(batch_images)
            batch_images_emb = F.normalize(image_features, dim=-1)
            batch_texts_emb = F.normalize(model.encode_text(batch_texts_tok), dim=-1)

        batch_images_emb_list.append(batch_images_emb.cpu())
        batch_texts_emb_list.append(batch_texts_emb.cpu())
        texts_image_index.extend(batch_texts_image_index)

    batch_size = len(batch_images_emb_list[0])

    # concatenate all embeddings
    images_emb = torch.cat(batch_images_emb_list)
    texts_emb = torch.cat(batch_texts_emb_list)

    # get the score for each text and image pair
    scores = texts_emb @ images_emb.t()

    scores_T = scores.T

    if args.reweight_retrieval:
        scores = scores * scores.softmax(dim=0)
        scores_T = scores_T * scores_T.softmax(dim=0)

    # construct a the positive pair matrix, which tells whether each text-image pair is a positive or not
    positive_pairs = torch.zeros_like(scores, dtype=bool)
    positive_pairs[torch.arange(len(scores)), texts_image_index] = True
    metrics = {}
    for recall_k in recall_k_list:
        # Note that recall_at_k computes **actual** recall i.e. nb_true_positive/nb_positives, where the number
        # of true positives, e.g. for text retrieval, is, for each image,  the number of retrieved texts matching that image among the top-k.
        # Also, the number of positives are the total number of texts matching the image in the dataset, as we have a set of captions
        # for each image, that number will be greater than 1 for text retrieval.
        # However, image/text retrieval recall@k, the way it is done in CLIP-like papers, is a bit different.
        # recall@k, in CLIP-like papers, is, for each image, either 1 or 0. It is 1 if atleast one text matches the image among the top-k.
        # so we can easily compute that using the actual recall, by checking whether there is at least one true positive,
        # which would be the case if the recall is greater than 0. One we compute the recal for each image (or text), we average
        # it over the dataset.
        metrics[f"image_retrieval_recall@{recall_k}"] = (
            (
                batchify(
                    recall_at_k, scores, positive_pairs, batch_size, device, k=recall_k
                )
                > 0
            )
            .float()
            .mean()
            .item()
        )
        metrics[f"text_retrieval_recall@{recall_k}"] = (
            (
                batchify(
                    recall_at_k,
                    scores_T,
                    positive_pairs.T,
                    batch_size,
                    device,
                    k=recall_k,
                )
                > 0
            )
            .float()
            .mean()
            .item()
        )

    return metrics


def dataloader_with_indices(dataloader):
    start = 0
    for x, y in dataloader:
        end = start + len(y)
        inds = torch.arange(start, end)
        yield x, y, inds
        start = end


def recall_at_k(scores, positive_pairs, k):
    """
    Compute the recall at k for each sample
    :param scores: compability score between  text and image embeddings (nb texts, nb images)
    :param k: number of images to consider per text, for retrieval
    :param positive_pairs: boolean matrix of positive pairs (nb texts, nb images)
    :return: recall at k averaged over all texts
    """
    nb_texts, nb_images = scores.shape
    # for each text, sort according to image scores in decreasing order
    topk_indices = torch.topk(scores, k, dim=1)[1]
    # compute number of positives for each text
    nb_positive = positive_pairs.sum(dim=1)
    # nb_texts, k, nb_images
    topk_indices_onehot = torch.nn.functional.one_hot(
        topk_indices, num_classes=nb_images
    )
    # compute number of true positives
    positive_pairs_reshaped = positive_pairs.view(nb_texts, 1, nb_images)
    # a true positive means a positive among the topk
    nb_true_positive = (topk_indices_onehot * positive_pairs_reshaped).sum(dim=(1, 2))
    # compute recall at k
    recall_at_k = nb_true_positive / nb_positive
    return recall_at_k


def batchify(func, X, Y, batch_size, device, *args, **kwargs):
    results = []
    for start in range(0, len(X), batch_size):
        end = start + batch_size
        x = X[start:end].to(device)
        y = Y[start:end].to(device)
        result = func(x, y, *args, **kwargs).cpu()
        results.append(result)
    return torch.cat(results)



"""Verify that METRICS.md covers all QualityMetrics output fields.

METRICS.md is auto-generated by ``ayase modules docs``. This test ensures
every module-written QualityMetrics field appears somewhere in the generated
doc, and every QualityMetrics field is accounted for.
"""

from pathlib import Path

from ayase.models import QualityMetrics


def _load_metrics_md_fields():
    """Extract all backtick-quoted field names from METRICS.md output columns."""
    metrics_path = Path("METRICS.md")
    if not metrics_path.exists():
        return set()
    text = metrics_path.read_text(encoding="utf-8")
    import re
    # Match field names in output column: `field_name` - description
    return set(re.findall(r"`(\w+)`\s*-\s*", text))


def test_metrics_table_matches_quality_metrics():
    """Verify METRICS.md references are consistent with QualityMetrics model."""
    fields_in_doc = _load_metrics_md_fields()
    model_fields = set(QualityMetrics.model_fields.keys())

    # Every field mentioned in METRICS.md must exist in QualityMetrics
    unknown = fields_in_doc - model_fields
    # Filter out non-field references (e.g. config keys)
    unknown = {f for f in unknown if not any(
        kw in f for kw in ("threshold", "model", "subsample", "device", "mode")
    )}
    assert not unknown, f"METRICS.md references unknown fields: {unknown}"

    # QualityMetrics field count must be correct
    assert len(model_fields) == 354, f"Expected 354 fields, got {len(model_fields)}"



"""Temporal Information (TI) and Spatial Information (SI) module.

ITU-T P.910 standard metrics used worldwide for characterising video
content complexity.

SI (Spatial Information):
  For each frame, apply a Sobel filter. SI is the maximum (over all
  frames) of the standard deviation of the filtered luminance frame.
  Higher SI means more spatial detail / edges.

TI (Temporal Information):
  For each pair of consecutive frames, compute the pixel-wise
  difference. TI is the maximum (over all frame-pairs) of the standard
  deviation of that difference frame.
  Higher TI means more temporal change / motion.

Both metrics are unit-less, typical ranges roughly 0-250 but depend on
content and resolution.  No external dependencies beyond OpenCV.
"""

import logging
from typing import Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class TISIModule(PipelineModule):
    name = "ti_si"
    description = "ITU-T P.910 Temporal & Spatial Information"
    default_config = {
        "max_frames": 300,  # Cap to avoid very long videos
    }

    def __init__(self, config=None):
        super().__init__(config)
        self.max_frames = self.config.get("max_frames", 300)

    def _compute_si(self, gray: np.ndarray) -> float:
        """Standard deviation of Sobel-filtered luminance frame."""
        sobel_x = cv2.Sobel(gray, cv2.CV_64F, 1, 0, ksize=3)
        sobel_y = cv2.Sobel(gray, cv2.CV_64F, 0, 1, ksize=3)
        sobel = np.hypot(sobel_x, sobel_y)
        return float(sobel.std())

    def _compute_image(self, sample: Sample) -> Sample:
        """SI only (no TI for a single image)."""
        img = cv2.imread(str(sample.path))
        if img is None:
            logger.warning(f"Cannot read image: {sample.path}")
            return sample

        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        si = self._compute_si(gray)

        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()

        sample.quality_metrics.spatial_information = si
        logger.debug(f"SI for {sample.path.name}: {si:.2f}")
        return sample

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return self._compute_image(sample)

        cap = cv2.VideoCapture(str(sample.path))
        if not cap.isOpened():
            logger.warning(f"Cannot open video: {sample.path}")
            return sample

        try:
            si_values = []
            ti_values = []
            prev_gray: Optional[np.ndarray] = None
            frame_idx = 0

            while frame_idx < self.max_frames:
                ret, frame = cap.read()
                if not ret:
                    break

                gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

                # --- SI ---
                si_values.append(self._compute_si(gray))

                # --- TI ---
                if prev_gray is not None:
                    diff = gray.astype(np.float64) - prev_gray.astype(np.float64)
                    ti_values.append(float(diff.std()))

                prev_gray = gray
                frame_idx += 1

            if not si_values:
                return sample

            # ITU-T P.910 defines SI/TI as the *maximum* over time, but the
            # mean is also commonly reported.  We store the max (standard).
            si = float(np.max(si_values))
            ti = float(np.max(ti_values)) if ti_values else 0.0

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            sample.quality_metrics.spatial_information = si
            sample.quality_metrics.temporal_information = ti

            logger.debug(
                f"TI/SI for {sample.path.name}: SI={si:.2f}, TI={ti:.2f} "
                f"({frame_idx} frames)"
            )

        except Exception as e:
            logger.error(f"TI/SI failed for {sample.path}: {e}")
        finally:
            cap.release()

        return sample



"""Tests for strred module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_strred_basics():
    from ayase.modules.strred import STRREDModule
    _test_module_basics(STRREDModule, "strred")

def test_strred_no_reference(image_sample):
    from ayase.modules.strred import STRREDModule
    m = STRREDModule()
    result = m.process(image_sample)
    assert result is image_sample



"""Tests for face_landmark_quality module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_face_landmark_quality_basics():
    from ayase.modules.face_landmark_quality import FaceLandmarkQualityModule
    _test_module_basics(FaceLandmarkQualityModule, "face_landmark_quality")

def test_face_landmark_quality_video(video_sample):
    from ayase.modules.face_landmark_quality import FaceLandmarkQualityModule
    video_sample.quality_metrics = QualityMetrics()
    m = FaceLandmarkQualityModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



import io
import logging
import pickle
from contextlib import suppress

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn.metrics import balanced_accuracy_score, classification_report
from tqdm import tqdm

# from open_clip import image_to_device


def evaluate(model, dataloader, device, visargs, amp=True, args=None):
    autocast = torch.cuda.amp.autocast if amp else suppress
    model.visual.output_tokens = True

    if visargs.delete_post_ln:
        model.visual.ln_post = nn.Identity()
    if visargs.extract_layer is not None and visargs.extract_layer != -1:
        model.visual.transformer.resblocks = model.visual.transformer.resblocks[
            : (visargs.extract_layer + 1)
        ]

    if visargs.attn_rollout is not None:
        for module in model.visual.transformer.resblocks:
            if hasattr(module, "attn"):
                module.attn.forward = module.attn.forward_with_attn

    with torch.no_grad():
        for images, target in dataloader:
            old_images = images
            # images = image_to_device(images, device, torch.float32, mean=args.image_mean, std=args.image_std)
            iamges = images.to(device)
            target = target.to(device)

            with autocast():
                # predict
                if hasattr(model.visual, "trunk"):
                    x = model.visual.trunk.forward_features(images)
                else:
                    x = model.visual(images).latent

                # apply attn pooler if we want
                if visargs.attn_pooling is not None:
                    idx = {"Q": 0, "K": 1, "V": 2}[visargs.attn_pooling]
                    attn_layer = model.visual.attn_pool.attn
                    mat = attn_layer.in_proj_weight[
                        attn_layer.embed_dim * idx : attn_layer.embed_dim * (idx + 1)
                    ]
                    x = x @ mat.T

                    if attn_layer.in_proj_bias is not None:
                        x = (
                            x
                            + attn_layer.in_proj_bias[
                                None,
                                None,
                                idx
                                * attn_layer.embed_dim : (idx + 1)
                                * attn_layer.embed_dim,
                            ]
                        )

                x = x.reshape(target.shape[0], -1, x.shape[-1])

                if isinstance(images, list):
                    old_images = old_images[0]

                if (
                    hasattr(model.visual, "embed_cls_token")
                    and model.visual.embed_cls_token
                ) or (
                    hasattr(model.visual, "trunk")
                    and hasattr(model.visual.trunk, "cls_token")
                    and model.visual.trunk.cls_token
                ):
                    if isinstance(images, list):
                        x = x[:, :-1]
                        old_images = old_images.reshape(
                            target.shape[0], -1, *old_images.shape[1:]
                        )
                        old_images = old_images[:, :-1]
                    else:
                        x = x[:, 1:]

                sidelen = int(x.shape[1] ** 0.5)

                if isinstance(images, list):
                    old_images = old_images.reshape(
                        target.shape[0], sidelen, sidelen, 3, *old_images.shape[-2:]
                    )
                    psz = old_images.shape[-1]
                    old_images = old_images.permute(0, 3, 1, 4, 2, 5).reshape(
                        target.shape[0], 3, sidelen * psz, sidelen * psz
                    )

                x = x.view(x.shape[0], sidelen, sidelen, -1)
                x = x.permute(0, 3, 1, 2).contiguous()
                # convert on umpy
                x = x.cpu().numpy().astype(np.float16)

                old_images = old_images.cpu().numpy().astype(np.float16)
                old_images *= np.array(args.image_std)[None, :, None, None]
                old_images += np.array(args.image_mean)[None, :, None, None]
                data = {"image": old_images, "features": x}

                if visargs.attn_pooling is not None:
                    data["attn"] = {
                        "qkv": attn_layer.in_proj_weight.data,
                        "bias": (
                            attn_layer.in_proj_bias.data
                            if attn_layer.in_proj_bias is not None
                            else None
                        ),
                        "probe": model.visual.attn_pool.probe,
                    }

                if visargs.attn_rollout:
                    attns = [
                        module.attn.attn.mean(dim=1)
                        for module in model.visual.transformer.resblocks
                        if hasattr(module, "attn")
                    ]

                    def roll(i, rollout=None):
                        for attn in attns[i:]:
                            rollout = (
                                attn.float() @ rollout
                                if rollout is not None
                                else attn.float()
                            )
                        return rollout

                    # rollout = torch.stack([roll(i) for i in range(len(attns))], dim=1)
                    rollout = torch.stack(attns, dim=1)
                    data["rollout"] = rollout.cpu().numpy().astype(np.float16)

                bytes_str = pickle.dumps(data)

                print(f"<<<<<<<<<<<<<")
                print(bytes_str)
                print(f">>>>>>>>>>>>>")

            break

    exit()
    return None



from PIL import Image
import os

import torch
from transformers import AutoTokenizer
from ...constants import HF_CACHE_DIR, IMAGE_TOKEN_INDEX

# from moviepy.editor import VideoFileClip


import cv2
import numpy as np
from math import ceil, sqrt

def extract_frames(video_path, num_frames, output_dir):
    # Create output directory if it doesn't exist
    os.makedirs(output_dir, exist_ok=True)
    
    # Open the video file
    video = cv2.VideoCapture(video_path)
    
    # Get total number of frames in the video
    total_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
    
    # Calculate the step size to get equally-spaced frames
    step = total_frames // num_frames
    
    for i in range(num_frames):
        # Set the frame position
        video.set(cv2.CAP_PROP_POS_FRAMES, i * step)
        
        # Read the frame
        ret, frame = video.read()
        
        if ret:
            # Save the frame as an image
            output_path = os.path.join(output_dir, f"frame_{i:04d}.jpg")
            cv2.imwrite(output_path, frame)
        else:
            print(f"Failed to extract frame {i}")
    
    # Release the video object
    video.release()

    print(f"Extracted {num_frames} frames from the video.")

# Following functions adapted from: https://github.com/mapluisch/LLaVA-CLI-with-multiple-images/blob/main/llava-multi-images.py
def concatenate_images_vertical(images, dist_images):
    # calc max width from imgs
    width = max(img.shape[1] for img in images)
    # calc total height of imgs + dist between them
    total_height = sum(img.shape[0] for img in images) + dist_images * (len(images) - 1)

    # create new img with calculated dimensions, black bg
    new_img = np.zeros((total_height, width, 3), dtype=np.uint8)

    # init var to track current height pos
    current_height = 0
    for img in images:
        h, w = img.shape[:2]
        # paste img in new_img at current height
        new_img[current_height:current_height+h, :w] = img
        # update current height for next img
        current_height += h + dist_images

    return new_img

def concatenate_images_horizontal(images, dist_images):
    # calc total width of imgs + dist between them
    total_width = sum(img.shape[1] for img in images) + dist_images * (len(images) - 1)
    # calc max height from imgs
    height = max(img.shape[0] for img in images)

    # create new img with calculated dimensions, black bg
    new_img = np.zeros((height, total_width, 3), dtype=np.uint8)

    # init var to track current width pos
    current_width = 0
    for img in images:
        h, w = img.shape[:2]
        # paste img in new_img at current width
        new_img[:h, current_width:current_width+w] = img
        # update current width for next img
        current_width += w + dist_images

    return new_img

def concatenate_images_grid(images, dist_images, output_size):
    num_images = len(images)
    # calc grid size based on amount of input imgs
    grid_size = max(2, ceil(sqrt(num_images)))

    cell_width = (output_size[0] - dist_images * (grid_size - 1)) // grid_size
    cell_height = (output_size[1] - dist_images * (grid_size - 1)) // grid_size

    # create new img with output_size, black bg
    new_img = np.zeros((output_size[1], output_size[0], 3), dtype=np.uint8)

    for index, img in enumerate(images):
        # calc img aspect ratio
        img_ratio = img.shape[1] / img.shape[0]
        # calc target aspect ratio per cell
        target_ratio = cell_width / cell_height

        # resize img to fit in cell
        if img_ratio > target_ratio:
            new_width = cell_width
            new_height = int(cell_width / img_ratio)
        else:
            new_width = int(cell_height * img_ratio)
            new_height = cell_height

        # resize img using inter_lanczos
        resized_img = cv2.resize(img, (new_width, new_height), interpolation=cv2.INTER_LANCZOS4)

        row = index // grid_size
        col = index % grid_size

        # calc x, y offsets for img positioning
        x_offset = col * (cell_width + dist_images) + (cell_width - new_width) // 2
        y_offset = row * (cell_height + dist_images) + (cell_height - new_height) // 2

        # paste resized img in calc pos
        new_img[y_offset:y_offset+new_height, x_offset:x_offset+new_width] = resized_img

    return new_img

def expand2square(pil_img, background_color):
    width, height = pil_img.size
    if width == height:
        return pil_img
    elif width > height:
        result = Image.new(pil_img.mode, (width, width), background_color)
        result.paste(pil_img, (0, (width - height) // 2))
        return result
    else:
        result = Image.new(pil_img.mode, (height, height), background_color)
        result.paste(pil_img, ((height - width) // 2, 0))
        return result


def tokenizer_image_token(prompt, tokenizer, image_token_index=IMAGE_TOKEN_INDEX, return_tensors=None):
    prompt_chunks = [tokenizer(chunk).input_ids for chunk in prompt.split('<image>')]

    def insert_separator(X, sep):
        return [ele for sublist in zip(X, [sep]*len(X)) for ele in sublist][:-1]

    input_ids = []
    offset = 0
    if len(prompt_chunks) > 0 and len(prompt_chunks[0]) > 0 and prompt_chunks[0][0] == tokenizer.bos_token_id:
        offset = 1
        input_ids.append(prompt_chunks[0][0])

    for x in insert_separator(prompt_chunks, [image_token_index] * (offset + 1)):
        input_ids.extend(x[offset:])

    if return_tensors is not None:
        if return_tensors == 'pt':
            return torch.tensor(input_ids, dtype=torch.long)
        raise ValueError(f'Unsupported tensor type: {return_tensors}')
    return input_ids


def t5_tokenizer_image_token(prompt, tokenizer, image_token_index=IMAGE_TOKEN_INDEX, return_tensors=None):
    prompt_chunks = [tokenizer(chunk).input_ids for chunk in prompt.split('<image>')]

    def insert_separator(X, sep):
        return [ele for sublist in zip(X, [sep]*len(X)) for ele in sublist][:-1]

    input_ids = []
    # Since there's no bos_token_id, simply concatenate the tokenized prompt_chunks with the image_token_index
    for x in insert_separator(prompt_chunks, [image_token_index]):
        input_ids.extend(x)

    if return_tensors is not None:
        if return_tensors == 'pt':
            return torch.tensor(input_ids, dtype=torch.long)
        raise ValueError(f'Unsupported tensor type: {return_tensors}')
    return input_ids


def load_pretrained_model(model_cls,
                          model_args,
                          model_path=None,
                          tokenizer_path=None,
                          model_max_length=None,
                          padding_side=None,
                          image_aspect_ratio='pad', # or 'square'
                          mmprojector_repo=None,
                          mmprojector_name=None,
                          device='cuda',
                          cache_dir=HF_CACHE_DIR):
    tokenizer_dict = {}
    if model_max_length:
        tokenizer_dict['model_max_length'] = model_max_length
    if padding_side:
        tokenizer_dict['padding_side'] = padding_side
    tokenizer = AutoTokenizer.from_pretrained(tokenizer_path, use_fast=False, **tokenizer_dict)
    # tokenizer.pad_token = tokenizer.unk_token # could be redundant

    model = model_cls.from_pretrained(model_path, cache_dir=cache_dir)
    
    if mmprojector_repo:
        from huggingface_hub import hf_hub_download
        model_base_name = mmprojector_repo.split('/')[-1]
        
        if cache_dir is not None:
            local_dir = os.path.join(cache_dir, model_base_name)
        elif os.environ.get('HF_HOME') is not None:
            local_dir = os.path.join(os.environ.get('HF_HOME'), model_base_name)
        else:
            local_dir = os.path.join(os.path.expanduser("~"), model_base_name)
        print(f"Downloading projector weights to {local_dir}")
        hf_hub_download(
            repo_id=mmprojector_repo,
            filename=mmprojector_name,
            local_dir=local_dir,
        )
        pretrain_mm_mlp_adapter = os.path.join(local_dir, mmprojector_name)
        model_args.pretrain_mm_mlp_adapter = pretrain_mm_mlp_adapter # important to set to correct path
        
        model.get_model().initialize_vision_modules(model_args) # This will load the CLIP vision encoder and MLP projector
    else:
        model.resize_token_embeddings(len(tokenizer)) # perhaps not needed

    if not model.get_vision_tower().is_loaded:
        model.get_vision_tower().load_model()
    model.to(device=device, dtype=torch.bfloat16)
    image_processor = model.get_vision_tower().image_processor

    model.requires_grad_(False)
    
    
    # below might be redundant
    model.config.image_aspect_ratio = image_aspect_ratio
    model.config.use_cache = False
    model.config.image_grid_pinpoints = None
    model.config.freeze_mm_mlp_adapter = True

    model = model.eval()
    return tokenizer, model, image_processor


"""Tests for playback_speed module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_playback_speed_basics():
    from ayase.modules.playback_speed import PlaybackSpeedModule
    _test_module_basics(PlaybackSpeedModule, "playback_speed")

def test_playback_speed_video(video_sample):
    from ayase.modules.playback_speed import PlaybackSpeedModule
    video_sample.quality_metrics = QualityMetrics()
    m = PlaybackSpeedModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""Tests for action_recognition module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_action_recognition_basics():
    from ayase.modules.action_recognition import ActionRecognitionModule
    _test_module_basics(ActionRecognitionModule, "action_recognition")

def test_action_recognition_video(video_sample):
    from ayase.modules.action_recognition import ActionRecognitionModule
    video_sample.quality_metrics = QualityMetrics()
    m = ActionRecognitionModule()
    m.on_mount()
    result = m.process(video_sample)
    assert result is video_sample



"""NR-GVQM — No-Reference Gaming Video Quality Metric (ISM 2018).

Nine frame-level features are extracted and combined via a weighted
SVR-like regression to predict perceptual quality of gaming/screen-
capture video content.

Features:
  1. Naturalness (NSS deviation from Gaussian)
  2. Blockiness (DCT block-boundary gradient ratio)
  3. Blur (Laplacian variance)
  4. Noise (high-frequency energy in smooth patches)
  5. Contrast (RMS contrast)
  6. Colorfulness (Hasler-Susstrunk metric)
  7. Brightness (mean luminance deviation from mid-grey)
  8. Edge density (Canny edge pixel ratio)
  9. Texture complexity (LBP uniformity proxy)

Output: ``gamival_score`` (reuses existing field since GAMIVAL
supersedes NR-GVQM for gaming VQA).
"""

import logging
from typing import List, Optional

import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
# Feature extraction
# ---------------------------------------------------------------------------

def _naturalness(gray: np.ndarray) -> float:
    """NSS regularity: deviation of MSCN from unit-variance Gaussian."""
    mu = cv2.GaussianBlur(gray, (7, 7), 7 / 6)
    sigma = np.sqrt(
        np.abs(cv2.GaussianBlur(gray * gray, (7, 7), 7 / 6) - mu * mu)
    )
    sigma = np.maximum(sigma, 1e-7)
    mscn = (gray - mu) / sigma

    # Kurtosis of a Gaussian is 3; deviation indicates non-naturalness
    kurtosis = float(np.mean(mscn ** 4) / (np.var(mscn) ** 2 + 1e-8))
    # Score: closer to 3 = more natural (higher quality)
    return 1.0 / (1.0 + abs(kurtosis - 3.0) * 0.5)


def _blockiness(gray: np.ndarray, block_size: int = 8) -> float:
    """Ratio of gradient at block boundaries vs interior (lower = better)."""
    h, w = gray.shape
    if h < block_size * 3 or w < block_size * 3:
        return 1.0  # Cannot reliably measure

    boundary_g = []
    interior_g = []

    for y in range(block_size, h - 1, block_size):
        # Measure gradient ACROSS the block boundary (row y-1 vs row y)
        g = np.mean(np.abs(gray[y, :].astype(np.float64) - gray[y - 1, :].astype(np.float64)))
        boundary_g.append(g)

    for y in range(1, h - 1):
        if y % block_size != 0:
            g = np.mean(np.abs(gray[y, :].astype(np.float64) - gray[y - 1, :].astype(np.float64)))
            interior_g.append(g)

    if not interior_g:
        return 1.0

    ratio = np.mean(boundary_g) / (np.mean(interior_g) + 1e-8)
    # Quality: ratio close to 1 = no blocking
    return 1.0 / (1.0 + max(0.0, ratio - 1.0) * 3.0)


def _blur(gray: np.ndarray) -> float:
    """Sharpness via Laplacian variance (higher = sharper = better quality)."""
    lap_var = cv2.Laplacian(gray, cv2.CV_64F).var()
    # Log scale normalisation; typical sharp: 500-3000
    return min(1.0, np.log1p(lap_var) / np.log1p(2000.0))


def _noise(gray: np.ndarray) -> float:
    """Estimate noise level from high-frequency energy in smooth patches."""
    h, w = gray.shape
    patch_size = 32
    noise_vals = []

    for y in range(0, h - patch_size, patch_size * 2):
        for x in range(0, w - patch_size, patch_size * 2):
            patch = gray[y:y + patch_size, x:x + patch_size]
            if np.std(patch) < 20:  # Smooth region
                # High-pass via Laplacian
                lap = cv2.Laplacian(patch, cv2.CV_64F)
                noise_vals.append(np.std(lap))

    if not noise_vals:
        return 0.8  # Assume moderate quality if no smooth patches found

    noise_level = np.mean(noise_vals)
    # Quality: lower noise = better (typical range 0-30)
    return 1.0 / (1.0 + noise_level * 0.1)


def _contrast(gray: np.ndarray) -> float:
    """RMS contrast (global standard deviation of luminance)."""
    c = float(np.std(gray))
    # Good contrast: 40-80; too low or too high is worse
    ideal = 60.0
    deviation = abs(c - ideal) / ideal
    return max(0.0, 1.0 - deviation)


def _colorfulness(frame: np.ndarray) -> float:
    """Hasler-Susstrunk colorfulness metric, normalised to 0-1."""
    b = frame[:, :, 0].astype(np.float64)
    g = frame[:, :, 1].astype(np.float64)
    r = frame[:, :, 2].astype(np.float64)

    rg = r - g
    yb = 0.5 * (r + g) - b
    cf = np.sqrt(rg.var() + yb.var()) + 0.3 * np.sqrt(rg.mean() ** 2 + yb.mean() ** 2)
    # Normalise: typical range 0-150, gaming content often 30-100
    return min(1.0, cf / 100.0)


def _brightness(gray: np.ndarray) -> float:
    """Mean brightness quality — penalise deviation from mid-grey."""
    mean_lum = np.mean(gray)
    # Ideal brightness ~120-140 for 8-bit
    deviation = abs(mean_lum - 128.0) / 128.0
    return max(0.0, 1.0 - deviation)


def _edge_density(gray: np.ndarray) -> float:
    """Fraction of pixels that are edges (Canny)."""
    edges = cv2.Canny(gray.astype(np.uint8), 50, 150)
    density = np.sum(edges > 0) / (edges.shape[0] * edges.shape[1])
    # Gaming content typically has moderate edge density: ~0.05-0.15
    # Too low = blurry, too high = noisy
    ideal = 0.10
    deviation = abs(density - ideal) / (ideal + 1e-8)
    return max(0.0, 1.0 - deviation * 0.5)


def _texture_complexity(gray: np.ndarray) -> float:
    """Texture complexity via gradient magnitude statistics (LBP proxy)."""
    gx = cv2.Sobel(gray, cv2.CV_64F, 1, 0, ksize=3)
    gy = cv2.Sobel(gray, cv2.CV_64F, 0, 1, ksize=3)
    grad_mag = np.sqrt(gx ** 2 + gy ** 2)

    # Entropy of gradient magnitude histogram
    hist, _ = np.histogram(grad_mag, bins=64, range=(0, np.max(grad_mag) + 1e-8))
    hist = hist.astype(np.float64)
    hist = hist / (hist.sum() + 1e-8)
    entropy = -np.sum(hist[hist > 0] * np.log2(hist[hist > 0]))

    # Normalise: max entropy for 64 bins = 6
    return min(1.0, entropy / 6.0)


def _extract_9_features(frame: np.ndarray) -> np.ndarray:
    """Extract the 9 NR-GVQM frame-level features."""
    gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)

    return np.array([
        _naturalness(gray),
        _blockiness(gray),
        _blur(gray),
        _noise(gray),
        _contrast(gray),
        _colorfulness(frame),
        _brightness(gray),
        _edge_density(gray),
        _texture_complexity(gray),
    ], dtype=np.float64)


# Heuristic SVR-like weights (learned from typical gaming VQA datasets)
_WEIGHTS = np.array([
    0.15,   # naturalness
    0.15,   # blockiness
    0.15,   # blur / sharpness
    0.12,   # noise
    0.10,   # contrast
    0.08,   # colorfulness
    0.08,   # brightness
    0.08,   # edge density
    0.09,   # texture complexity
], dtype=np.float64)


class NRGVQMModule(PipelineModule):
    name = "nr_gvqm"
    description = "NR-GVQM no-reference gaming video quality (ISM 2018, 9 features)"
    default_config = {
        "subsample": 8,
    }

    def __init__(self, config=None):
        super().__init__(config)
        self._model = None
        self._backend = "heuristic"

    def setup(self) -> None:
        # NR-GVQM is a heuristic-only module (no official package exists)
        self._backend = "heuristic"
        logger.info("NR-GVQM initialised (heuristic 9-feature model)")

    def process(self, sample: Sample) -> Sample:
        try:
            score = self._compute_quality(sample)
            if score is not None:
                if sample.quality_metrics is None:
                    sample.quality_metrics = QualityMetrics()
                sample.quality_metrics.nr_gvqm_score = score
        except Exception as e:
            logger.warning("NR-GVQM failed for %s: %s", sample.path, e)
        return sample

    def _compute_quality(self, sample: Sample) -> Optional[float]:
        subsample = self.config.get("subsample", 8)

        if sample.is_video:
            return self._process_video(sample, subsample)
        else:
            return self._process_image(sample)

    def _process_image(self, sample: Sample) -> Optional[float]:
        """Evaluate a single image."""
        frame = cv2.imread(str(sample.path))
        if frame is None:
            return None
        features = _extract_9_features(frame)
        return self._predict(features)

    def _process_video(self, sample: Sample, subsample: int) -> Optional[float]:
        """Evaluate a video: extract features per frame, aggregate, predict."""
        cap = cv2.VideoCapture(str(sample.path))
        try:
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                return None

            indices = np.linspace(0, total - 1, min(subsample, total), dtype=int)
            all_features: List[np.ndarray] = []

            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    all_features.append(_extract_9_features(frame))
        finally:
            cap.release()

        if not all_features:
            return None

        # Average features across frames
        avg_features = np.mean(all_features, axis=0)

        # Temporal consistency penalty: high variance across frames is bad
        if len(all_features) > 1:
            feat_stack = np.array(all_features)
            temporal_var = np.mean(np.var(feat_stack, axis=0))
            temporal_penalty = temporal_var * 0.5
        else:
            temporal_penalty = 0.0

        score = self._predict(avg_features) - temporal_penalty
        return float(np.clip(score, 0.0, 1.0))

    def _predict(self, features: np.ndarray) -> float:
        """Weighted combination of 9 features (SVR-like heuristic)."""
        score = float(np.dot(features, _WEIGHTS))
        return float(np.clip(score, 0.0, 1.0))



"""AIGC-VQA — Holistic Perception for AIGC Video Quality (CVPRW 2024).

3-branch: technical (3D-Swin) + aesthetic (ConvNext) + text-video alignment (BLIP).

aigcvqa_technical, aigcvqa_aesthetic, aigcvqa_alignment
"""

import logging
import cv2
import numpy as np

from ayase.models import Sample, QualityMetrics
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class AIGCVQAModule(PipelineModule):
    name = "aigcvqa"
    description = "AIGC-VQA holistic 3-branch AIGC perception (CVPRW 2024)"
    default_config = {"subsample": 8}

    def __init__(self, config=None):
        super().__init__(config)
        self.subsample = self.config.get("subsample", 8)
        self._backend = "heuristic"

    def setup(self) -> None:
        try:
            import aigcvqa
            self._model = aigcvqa
            self._backend = "native"
            logger.info("AIGC-VQA (native) initialised")
            return
        except ImportError:
            pass
        self._backend = "heuristic"
        logger.info("AIGC-VQA (heuristic)")

    def process(self, sample: Sample) -> Sample:
        try:
            frames = self._extract_frames(sample)
            if not frames:
                return sample

            if sample.quality_metrics is None:
                sample.quality_metrics = QualityMetrics()

            # Technical branch
            tech_scores = []
            for f in frames:
                gray = cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float64)
                sharpness = min(cv2.Laplacian(gray, cv2.CV_64F).var() / 500.0, 1.0)
                contrast = min(gray.std() / 65.0, 1.0)
                tech_scores.append(0.6 * sharpness + 0.4 * contrast)
            sample.quality_metrics.aigcvqa_technical = float(np.clip(np.mean(tech_scores), 0, 1))

            # Aesthetic branch
            aes_scores = []
            for f in frames:
                hsv = cv2.cvtColor(f, cv2.COLOR_BGR2HSV)
                sat = min(hsv[:, :, 1].astype(float).mean() / 128.0, 1.0)
                gray = cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(float)
                brightness = 1.0 - abs(gray.mean() - 127.5) / 127.5
                aes_scores.append(0.5 * sat + 0.5 * brightness)
            sample.quality_metrics.aigcvqa_aesthetic = float(np.clip(np.mean(aes_scores), 0, 1))

            # Alignment branch (proxy: caption presence check)
            caption = getattr(sample, "caption", None)
            if caption and hasattr(caption, "text") and caption.text:
                sample.quality_metrics.aigcvqa_alignment = 0.7  # Default when caption exists
            else:
                sample.quality_metrics.aigcvqa_alignment = 0.5  # No caption to align

        except Exception as e:
            logger.warning(f"AIGC-VQA failed for {sample.path}: {e}")
        return sample

    def _extract_frames(self, sample):
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            if total <= 0:
                cap.release()
                return []
            indices = np.linspace(0, total - 1, min(self.subsample, total), dtype=int)
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            img = cv2.imread(str(sample.path))
            if img is not None:
                frames.append(img)
        return frames



"""Tests for erqa module."""

from ..conftest import _test_module_basics
from ayase.models import QualityMetrics


def test_erqa_basics():
    from ayase.modules.erqa import ERQAModule
    _test_module_basics(ERQAModule, "erqa")

def test_erqa_no_reference(image_sample):
    from ayase.modules.erqa import ERQAModule
    m = ERQAModule()
    result = m.process(image_sample)
    assert result is image_sample



import cv2
import numpy as np

from ayase.models import QualityMetrics, Sample


def test_scene_detection_basics():
    from ayase.modules.scene_detection import SceneDetectionModule
    from .conftest import _test_module_basics

    _test_module_basics(SceneDetectionModule, "scene_detection")


def test_scene_detection_video(video_sample):
    from ayase.modules.scene_detection import SceneDetectionModule

    video_sample.quality_metrics = QualityMetrics()
    m = SceneDetectionModule()
    m.setup()
    result = m.process(video_sample)
    qm = result.quality_metrics
    assert qm is not None
    assert qm.scene_stability is not None
    assert 0 <= qm.scene_stability <= 1
    assert qm.avg_scene_duration is not None
    assert qm.avg_scene_duration > 0


def test_raft_motion_basics():
    from ayase.modules.raft_motion import RAFTMotionModule
    from .conftest import _test_module_basics

    _test_module_basics(RAFTMotionModule, "raft_motion")


def test_raft_motion_disabled(video_sample):
    from ayase.modules.raft_motion import RAFTMotionModule

    m = RAFTMotionModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_ram_tagging_basics():
    from ayase.modules.ram_tagging import RAMTaggingModule
    from .conftest import _test_module_basics

    _test_module_basics(RAMTaggingModule, "ram_tagging")


def test_ram_tagging_disabled(image_sample):
    from ayase.modules.ram_tagging import RAMTaggingModule

    m = RAMTaggingModule({})
    result = m.process(image_sample)
    assert result is image_sample


def test_depth_anything_basics():
    from ayase.modules.depth_anything import DepthAnythingModule
    from .conftest import _test_module_basics

    _test_module_basics(DepthAnythingModule, "depth_anything")


def test_depth_anything_disabled(video_sample):
    from ayase.modules.depth_anything import DepthAnythingModule

    m = DepthAnythingModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_video_type_classifier_basics():
    from ayase.modules.video_type_classifier import VideoTypeClassifierModule
    from .conftest import _test_module_basics

    _test_module_basics(VideoTypeClassifierModule, "video_type_classifier")


def test_video_type_classifier_disabled(video_sample):
    from ayase.modules.video_type_classifier import VideoTypeClassifierModule

    m = VideoTypeClassifierModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_jedi_basics():
    from ayase.modules.jedi_metric import JEDiModule
    from .conftest import _test_module_basics

    _test_module_basics(JEDiModule, "jedi")


def test_jedi_disabled(video_sample):
    from ayase.modules.jedi_metric import JEDiModule

    m = JEDiModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_trajan_basics():
    from ayase.modules.trajan import TRAJANModule
    from .conftest import _test_module_basics

    _test_module_basics(TRAJANModule, "trajan")


def test_trajan_video(video_sample):
    from ayase.modules.trajan import TRAJANModule

    video_sample.quality_metrics = QualityMetrics()
    m = TRAJANModule()
    m.setup()
    result = m.process(video_sample)
    qm = result.quality_metrics
    assert qm is not None
    assert qm.trajan_score is not None
    assert 0 <= qm.trajan_score <= 1


def test_trajan_image(image_sample):
    from ayase.modules.trajan import TRAJANModule

    m = TRAJANModule()
    m.setup()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.trajan_score is None


def test_promptiqa_basics():
    from ayase.modules.promptiqa import PromptIQAModule
    from .conftest import _test_module_basics

    _test_module_basics(PromptIQAModule, "promptiqa")


def test_promptiqa_disabled(image_sample):
    from ayase.modules.promptiqa import PromptIQAModule

    m = PromptIQAModule({})
    result = m.process(image_sample)
    assert result is image_sample


def test_aigv_assessor_basics():
    from ayase.modules.aigv_assessor import AIGVAssessorModule
    from .conftest import _test_module_basics

    _test_module_basics(AIGVAssessorModule, "aigv_assessor")


def test_aigv_assessor_video(video_sample):
    from ayase.modules.aigv_assessor import AIGVAssessorModule

    video_sample.quality_metrics = QualityMetrics()
    m = AIGVAssessorModule()
    m.setup()
    result = m.process(video_sample)
    qm = result.quality_metrics
    assert qm is not None
    assert qm.aigv_static is not None
    assert 0 <= qm.aigv_static <= 1
    assert qm.aigv_temporal is not None
    assert qm.aigv_dynamic is not None


def test_aigv_assessor_image(image_sample):
    from ayase.modules.aigv_assessor import AIGVAssessorModule

    m = AIGVAssessorModule()
    m.setup()
    result = m.process(image_sample)
    assert result.quality_metrics is None or result.quality_metrics.aigv_static is None


def test_video_reward_basics():
    from ayase.modules.video_reward import VideoRewardModule
    from .conftest import _test_module_basics

    _test_module_basics(VideoRewardModule, "video_reward")


def test_video_reward_disabled(video_sample):
    from ayase.modules.video_reward import VideoRewardModule

    m = VideoRewardModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_text_overlay_basics():
    from ayase.modules.text_overlay import TextOverlayModule
    from .conftest import _test_module_basics

    _test_module_basics(TextOverlayModule, "text_overlay")


def test_text_overlay_image(image_sample):
    from ayase.modules.text_overlay import TextOverlayModule

    image_sample.quality_metrics = QualityMetrics()
    m = TextOverlayModule()
    m.setup()
    result = m.process(image_sample)
    qm = result.quality_metrics
    assert qm is not None
    assert qm.text_overlay_score is not None
    assert 0 <= qm.text_overlay_score <= 1


def test_text_overlay_video(video_sample):
    from ayase.modules.text_overlay import TextOverlayModule

    video_sample.quality_metrics = QualityMetrics()
    m = TextOverlayModule()
    m.setup()
    result = m.process(video_sample)
    qm = result.quality_metrics
    assert qm is not None
    assert qm.text_overlay_score is not None
    assert 0 <= qm.text_overlay_score <= 1


def test_ptlflow_motion_basics():
    from ayase.modules.ptlflow_motion import PtlflowMotionModule
    from .conftest import _test_module_basics

    _test_module_basics(PtlflowMotionModule, "ptlflow_motion")


def test_ptlflow_motion_disabled(video_sample):
    from ayase.modules.ptlflow_motion import PtlflowMotionModule

    m = PtlflowMotionModule({})
    result = m.process(video_sample)
    assert result is video_sample


def test_qcn_basics():
    from ayase.modules.qcn import QCNModule
    from .conftest import _test_module_basics

    _test_module_basics(QCNModule, "qcn")


def test_qcn_disabled(image_sample):
    from ayase.modules.qcn import QCNModule

    m = QCNModule({})
    result = m.process(image_sample)
    assert result is image_sample


def test_motion_scene_qualitymetrics_fields():
    qm = QualityMetrics()
    fields = [
        "scene_stability",
        "avg_scene_duration",
        "raft_motion_score",
        "ptlflow_motion_score",
        "ram_tags",
        "depth_anything_score",
        "depth_anything_consistency",
        "video_type",
        "video_type_confidence",
        "trajan_score",
        "promptiqa_score",
        "qcn_score",
        "aigv_static",
        "aigv_temporal",
        "aigv_dynamic",
        "aigv_alignment",
        "video_reward_score",
        "text_overlay_score",
    ]
    for field in fields:
        assert hasattr(qm, field)


def test_jedi_dataset_stats_field():
    from ayase.models import DatasetStats

    stats = DatasetStats(
        total_samples=10,
        valid_samples=10,
        invalid_samples=0,
        total_size=1000)
    assert hasattr(stats, "jedi")
    assert stats.jedi is None


def test_temporal_flickering_load_all_frames_respects_max(tmp_path):
    """_load_all_frames subsamples when frame count exceeds max_frames."""
    from ayase.modules.temporal_flickering import TemporalFlickeringModule

    # Create a video with 30 frames
    vid = tmp_path / "long.mp4"
    fourcc = cv2.VideoWriter_fourcc(*"mp4v")
    writer = cv2.VideoWriter(str(vid), fourcc, 30.0, (32, 32))
    for _ in range(30):
        writer.write(np.full((32, 32, 3), 100, dtype=np.uint8))
    writer.release()

    module = TemporalFlickeringModule(config={"max_frames": 10})
    sample = Sample(path=vid, is_video=True)
    frames = module._load_all_frames(sample)

    assert len(frames) <= 10



"""C3DVQA (3D CNN Video Quality Assessment) module.

2020 — Uses 3D convolutions over spatiotemporal video volumes
to capture both spatial distortions and temporal artefacts in
a single forward pass. Operates on short video clips and pools
predictions across the full video.

This implementation uses a lightweight 3D feature extractor
with temporal convolutions for quality prediction.
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class C3DVQAModule(PipelineModule):
    name = "c3dvqa"
    description = "3D CNN spatiotemporal video quality assessment"
    default_config = {
        "clip_length": 16,
        "subsample": 4,
    }

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._ml_available = False
        self._model = None

    def setup(self) -> None:
        try:
            import torch
            import torchvision.models.video as video_models

            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            # Use R3D-18 as 3D feature extractor with pretrained weights
            try:
                weights = video_models.R3D_18_Weights.DEFAULT
            except AttributeError:
                weights = "DEFAULT"
            model = video_models.r3d_18(weights=weights)
            # Remove classifier, use as feature extractor
            self._feature_extractor = torch.nn.Sequential(
                model.stem,
                model.layer1,
                model.layer2,
                model.layer3,
                model.layer4,
                torch.nn.AdaptiveAvgPool3d(1),
            ).to(device)
            self._feature_extractor.eval()
            self._device = device
            self._ml_available = True
            logger.info("C3DVQA initialised (R3D-18 backbone) on %s", device)
        except (ImportError, Exception) as e:
            logger.info("C3DVQA ML backbone unavailable, using handcrafted 3D features: %s", e)
            self._ml_available = False

    def process(self, sample: Sample) -> Sample:
        if not sample.is_video:
            return sample
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        try:
            import cv2

            clip_length = self.config.get("clip_length", 16)
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

            # Sample clips uniformly
            n_clips = self.config.get("subsample", 4)
            clip_starts = np.linspace(0, max(0, total - clip_length), n_clips, dtype=int)

            clip_scores = []
            for start in clip_starts:
                frames = []
                for j in range(clip_length):
                    idx = min(start + j, total - 1)
                    cap.set(cv2.CAP_PROP_POS_FRAMES, int(idx))
                    ret, frame = cap.read()
                    if ret:
                        frames.append(frame)

                if len(frames) < 4:
                    continue

                score = self._score_clip(frames)
                clip_scores.append(score)

            cap.release()

            if clip_scores:
                # Percentile pooling (worst clips matter more)
                sorted_scores = sorted(clip_scores)
                n = len(sorted_scores)
                p10 = sorted_scores[max(0, int(n * 0.1))]
                final = 0.3 * p10 + 0.5 * np.mean(clip_scores) + 0.2 * np.median(clip_scores)
                sample.quality_metrics.c3dvqa_score = float(np.clip(final, 0.0, 1.0))
        except Exception as e:
            logger.warning("C3DVQA failed: %s", e)
        return sample

    def _score_clip(self, frames) -> float:
        """Score a clip of frames using 3D features."""
        if self._ml_available:
            return self._score_clip_ml(frames)
        return self._score_clip_handcrafted(frames)

    def _score_clip_ml(self, frames) -> float:
        """Score using 3D CNN features."""
        import torch
        import cv2

        target_size = (112, 112)
        processed = []
        for f in frames:
            rgb = cv2.cvtColor(f, cv2.COLOR_BGR2RGB)
            resized = cv2.resize(rgb, target_size)
            tensor = torch.from_numpy(resized).float() / 255.0
            processed.append(tensor)

        # Stack: (T, H, W, C) -> (C, T, H, W)
        clip = torch.stack(processed).permute(3, 0, 1, 2).unsqueeze(0)
        clip = clip.to(self._device)

        with torch.no_grad():
            features = self._feature_extractor(clip)
            features = features.squeeze()

        # Heuristic proxy: no trained regression head is available, so we
        # map raw feature statistics to [0,1].  This is an *experimental*
        # approximation — values are not calibrated to subjective MOS.
        # Note: heuristic proxy using raw feature statistics; not calibrated against subjective quality scores.
        feat_mean = features.mean().item()
        feat_std = features.std().item()
        activation_proxy = min(1.0, abs(feat_mean) / 2.0)
        diversity_proxy = min(1.0, feat_std / 1.0)

        return 0.5 * activation_proxy + 0.5 * diversity_proxy

    def _score_clip_handcrafted(self, frames) -> float:
        """Score using handcrafted 3D spatiotemporal features."""
        import cv2

        grays = [cv2.cvtColor(f, cv2.COLOR_BGR2GRAY).astype(np.float64) for f in frames]

        # Spatial quality: average frame sharpness
        sharpness_scores = []
        for g in grays:
            lap_var = cv2.Laplacian(g, cv2.CV_64F).var()
            sharpness_scores.append(min(1.0, lap_var / 500.0))

        # Temporal quality: 3D gradient (spatiotemporal edges)
        temporal_scores = []
        for i in range(1, len(grays)):
            # Temporal derivative
            dt = grays[i] - grays[i - 1]
            # Spatial gradients at current frame
            dx = cv2.Sobel(grays[i], cv2.CV_64F, 1, 0, ksize=3)
            dy = cv2.Sobel(grays[i], cv2.CV_64F, 0, 1, ksize=3)

            # 3D gradient magnitude
            grad_3d = np.sqrt(dx ** 2 + dy ** 2 + dt ** 2)
            # Coherent 3D structure suggests natural content
            coherence = np.mean(grad_3d) / (np.std(grad_3d) + 1e-8)
            temporal_scores.append(min(1.0, coherence / 2.0))

        spatial = float(np.mean(sharpness_scores))
        temporal = float(np.mean(temporal_scores)) if temporal_scores else 0.8

        return 0.5 * spatial + 0.5 * temporal



from .conv_backbone import convnext_3d_small, convnext_3d_tiny
from .evaluator import DOVER, BaseEvaluator, BaseImageEvaluator
from .head import IQAHead, VARHead, VQAHead
from .swin_backbone import SwinTransformer2D as IQABackbone
from .swin_backbone import SwinTransformer3D as VQABackbone
from .swin_backbone import swin_3d_small, swin_3d_tiny

__all__ = [
    "VQABackbone",
    "IQABackbone",
    "VQAHead",
    "IQAHead",
    "VARHead",
    "BaseEvaluator",
    "BaseImageEvaluator",
    "DOVER",
]



"""VIDEVAL (Video quality EVALuator) module.

Feature-fusion NR-VQA that extracts spatial and temporal features inspired
by BRISQUE, NIQE, GMLOG, HIGRADE, and FRIQUEE, and combines them via
a regression model.

Backend tiers:
  1. **VIDEVAL SVR** — pre-trained SVR from ``github.com/vztu/VIDEVAL``
     with expanded 60-feature extraction
  2. **Reduced features** — 24-feature extraction with heuristic linear weights
"""

import logging
from typing import Optional

import numpy as np

from ayase.models import QualityMetrics, Sample
from ayase.pipeline import PipelineModule

logger = logging.getLogger(__name__)


class VIDEVALModule(PipelineModule):
    name = "videval"
    description = "Feature-fusion NR-VQA (VIDEVAL-style SVR or heuristic linear mapping)"
    default_config = {"subsample": 8}

    def __init__(self, config: Optional[dict] = None) -> None:
        super().__init__(config)
        self._svr_model = None
        self._scaler = None
        self._backend = "heuristic"
        self.mos_min = self.config.get("mos_min", 1.0)
        self.mos_max = self.config.get("mos_max", 5.0)

    def setup(self) -> None:
        # Tier 1: Try loading pre-trained SVR model
        try:
            import joblib
            from pathlib import Path

            models_dir = Path(self.config.get("models_dir", "models")) / "videval"
            svr_path = models_dir / "videval_svr.pkl"
            scaler_path = models_dir / "videval_scaler.pkl"

            missing = []
            if not svr_path.exists():
                missing.append(str(svr_path))
            if not scaler_path.exists():
                missing.append(str(scaler_path))

            if not missing:
                self._svr_model = joblib.load(svr_path)
                self._scaler = joblib.load(scaler_path)
                self._backend = "svr"
                logger.info("VIDEVAL loaded pre-trained SVR model from %s", models_dir)
                return

            logger.warning(
                "VIDEVAL SVR model files not found: %s. "
                "To enable the ML tier, obtain 'videval_svr.pkl' and "
                "'videval_scaler.pkl' from the VIDEVAL repository "
                "(https://github.com/vztu/VIDEVAL) and place them in '%s'. "
                "Falling back to heuristic feature mapping.",
                ", ".join(missing),
                models_dir,
            )
        except ImportError:
            logger.warning(
                "VIDEVAL SVR requires the 'joblib' package (pip install joblib). "
                "Falling back to heuristic feature mapping."
            )
        except Exception as e:
            logger.warning("VIDEVAL SVR loading failed: %s. "
                           "Falling back to heuristic feature mapping.", e)

        # Tier 2: Heuristic linear combination
        self._backend = "heuristic"
        logger.info("VIDEVAL using heuristic feature mapping")

    def process(self, sample: Sample) -> Sample:
        if sample.quality_metrics is None:
            sample.quality_metrics = QualityMetrics()
        try:
            import cv2

            frames = self._load_frames(sample)
            if not frames:
                return sample

            all_features = []
            for frame in frames:
                feats = self._extract_features(frame)
                all_features.append(feats)

            avg_feats = np.mean(all_features, axis=0)
            temporal_feats = self._extract_temporal_features(frames)
            combined = np.concatenate([avg_feats, temporal_feats])

            if self._backend == "svr" and self._svr_model is not None:
                score = self._svr_predict(combined)
            else:
                score = self._feature_to_quality(combined)

            sample.quality_metrics.videval_score = float(np.clip(score, 0.0, 1.0))
        except Exception as e:
            logger.warning("VIDEVAL failed: %s", e)
        return sample

    def _svr_predict(self, features: np.ndarray) -> float:
        """Predict quality using the pre-trained SVR model."""
        feat_2d = features.reshape(1, -1)
        if self._scaler is not None:
            feat_2d = self._scaler.transform(feat_2d)
        prediction = self._svr_model.predict(feat_2d)[0]
        # SVR output may be in MOS range, normalize to 0-1
        return float((prediction - self.mos_min) / (self.mos_max - self.mos_min))

    def _extract_features(self, frame) -> np.ndarray:
        """Extract spatial quality features from a single frame."""
        import cv2

        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY).astype(np.float64)
        h, w = gray.shape

        features = []

        # 1-2: BRISQUE-like MSCN statistics
        mu = cv2.GaussianBlur(gray, (7, 7), 7 / 6)
        sigma = cv2.GaussianBlur((gray - mu) ** 2, (7, 7), 7 / 6)
        sigma = np.sqrt(sigma + 1e-7)
        mscn = (gray - mu) / (sigma + 1.0)
        features.extend([np.mean(mscn), np.std(mscn)])

        # 3-4: Horizontal/vertical pairwise MSCN products
        h_prod = mscn[:, :-1] * mscn[:, 1:]
        v_prod = mscn[:-1, :] * mscn[1:, :]
        features.extend([np.mean(h_prod), np.mean(v_prod)])

        # 5-6: Laplacian statistics (sharpness)
        lap = cv2.Laplacian(gray, cv2.CV_64F)
        features.extend([np.mean(np.abs(lap)), np.var(lap)])

        # 7-8: Gradient magnitude statistics (GMLOG-like)
        gx = cv2.Sobel(gray, cv2.CV_64F, 1, 0, ksize=3)
        gy = cv2.Sobel(gray, cv2.CV_64F, 0, 1, ksize=3)
        grad_mag = np.sqrt(gx ** 2 + gy ** 2)
        log_grad = np.log1p(grad_mag)
        features.extend([np.mean(log_grad), np.std(log_grad)])

        # 9-10: Color statistics (HIGRADE-like)
        hsv = cv2.cvtColor(frame, cv2.COLOR_BGR2HSV).astype(np.float64)
        features.extend([np.std(hsv[:, :, 1]), np.mean(hsv[:, :, 2])])

        # 11-12: DCT statistics (FRIQUEE-like)
        block = gray[:min(h, 256), :min(w, 256)]
        dct = cv2.dct(block.astype(np.float32))
        features.extend([np.mean(np.abs(dct)), np.std(dct)])

        # 13-14: NIQE-like NSS shape parameters
        features.extend([float(np.mean(np.abs(mscn))), float(np.var(mscn ** 2))])

        # 15-16: Entropy features
        hist = cv2.calcHist([gray.astype(np.uint8)], [0], None, [256], [0, 256]).flatten()
        hist = hist / (hist.sum() + 1e-8)
        entropy = -np.sum(hist[hist > 0] * np.log2(hist[hist > 0]))
        features.extend([entropy, np.std(hist)])

        # 17-20: Extended BRISQUE (diagonal MSCN products, kurtosis)
        d1_prod = mscn[:-1, :-1] * mscn[1:, 1:]
        d2_prod = mscn[:-1, 1:] * mscn[1:, :-1]
        features.extend([np.mean(d1_prod), np.mean(d2_prod)])
        mscn_kurtosis = float(np.mean(mscn ** 4) / (np.mean(mscn ** 2) ** 2 + 1e-8))
        mscn_skewness = float(np.mean(mscn ** 3) / (np.std(mscn) ** 3 + 1e-8))
        features.extend([mscn_kurtosis, mscn_skewness])

        # 21-24: Multi-scale Laplacian statistics
        current = gray.copy()
        for s in range(2):
            current = cv2.resize(current, (max(1, current.shape[1] // 2),
                                           max(1, current.shape[0] // 2)))
            lap_s = cv2.Laplacian(current, cv2.CV_64F)
            features.extend([np.mean(np.abs(lap_s)), np.var(lap_s)])

        return np.array(features, dtype=np.float64)

    def _extract_temporal_features(self, frames) -> np.ndarray:
        """Extract temporal features across frames."""
        import cv2

        features = []
        if len(frames) < 2:
            return np.zeros(12, dtype=np.float64)

        frame_diffs = []
        flow_mags = []
        flow_stds = []
        for i in range(len(frames) - 1):
            g1 = cv2.cvtColor(frames[i], cv2.COLOR_BGR2GRAY).astype(np.float64)
            g2 = cv2.cvtColor(frames[i + 1], cv2.COLOR_BGR2GRAY).astype(np.float64)

            diff = np.mean(np.abs(g1 - g2))
            frame_diffs.append(diff)

            flow = cv2.calcOpticalFlowFarneback(
                g1.astype(np.uint8), g2.astype(np.uint8),
                None, 0.5, 3, 15, 3, 5, 1.2, 0
            )
            mag = np.sqrt(flow[..., 0] ** 2 + flow[..., 1] ** 2)
            flow_mags.append(np.mean(mag))
            flow_stds.append(np.std(mag))

        # 1-4: Frame difference statistics
        features.extend([
            np.mean(frame_diffs), np.std(frame_diffs),
            np.min(frame_diffs), np.max(frame_diffs),
        ])

        # 5-8: Flow magnitude statistics
        features.extend([
            np.mean(flow_mags), np.std(flow_mags),
            np.min(flow_mags), np.max(flow_mags),
        ])

        # 9-12: Flow spatial std statistics
        features.extend([
            np.mean(flow_stds), np.std(flow_stds),
            np.min(flow_stds), np.max(flow_stds),
        ])

        return np.array(features, dtype=np.float64)

    def _feature_to_quality(self, features: np.ndarray) -> float:
        """Map feature vector to quality score via heuristic weights."""
        f = features.copy()
        f_min, f_max = f.min(), f.max()
        if f_max > f_min:
            f = (f - f_min) / (f_max - f_min)

        n_spatial = min(28, len(f))
        n_temporal = len(f) - n_spatial
        spatial = f[:n_spatial]
        temporal = f[n_spatial:]

        # Sharpness features (higher = better)
        sharpness = min(1.0, (spatial[4] + spatial[5]) / 2.0) if len(spatial) > 5 else 0.5

        # Noise features
        noise_free = 1.0 - min(1.0, abs(spatial[1]) * 2.0) if len(spatial) > 1 else 0.5

        # Gradient richness
        gradient = min(1.0, spatial[6]) if len(spatial) > 6 else 0.5

        # Color vibrancy
        color = min(1.0, spatial[8]) if len(spatial) > 8 else 0.5

        # Entropy
        info_content = min(1.0, spatial[14]) if len(spatial) > 14 else 0.5

        # Temporal stability
        if n_temporal >= 12:
            temporal_stability = 1.0 - min(1.0, temporal[1])
            motion_quality = float(np.exp(-0.5 * ((temporal[4] - 0.5) / 0.3) ** 2))
        elif n_temporal >= 8:
            temporal_stability = 1.0 - min(1.0, temporal[1])
            motion_quality = float(np.exp(-0.5 * ((temporal[4] - 0.5) / 0.3) ** 2))
        else:
            temporal_stability = 0.8
            motion_quality = 0.8

        score = (0.25 * sharpness + 0.15 * noise_free + 0.15 * gradient +
                 0.10 * color + 0.10 * info_content +
                 0.15 * temporal_stability + 0.10 * motion_quality)

        return float(score)

    def _load_frames(self, sample: Sample) -> list:
        import cv2

        subsample = self.config.get("subsample", 8)
        frames = []
        if sample.is_video:
            cap = cv2.VideoCapture(str(sample.path))
            total = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
            indices = list(range(0, total, max(1, total // subsample)))[:subsample]
            for idx in indices:
                cap.set(cv2.CAP_PROP_POS_FRAMES, idx)
                ret, frame = cap.read()
                if ret:
                    frames.append(frame)
            cap.release()
        else:
            frame = cv2.imread(str(sample.path))
            if frame is not None:
                frames.append(frame)
        return frames
